import os
import time

import docker


class DockerClient:
    """
    Get Information for the local docker instances
    """

    def __init__(self):
        # TODO os env and add trys around the rest when wassetn posible to create a docker client

        try:
            self.client = docker.APIClient(base_url="unix://var/run/docker.sock")
            self.containers = self.client.containers()
        except docker.errors.DockerException as e:
            print(e)
            self.client = None
            self.containers = None
        self.container_statuses = []
        self.sleep_time_for_cpu_usage_check = 3

    def get_stats_all(self):
        """
        Returns a list of dicts, of  all the memory usage and the cpu utilization for all locally running docker
        containers
        """
        container_statuses_start = self._get_all_stats_all_container()
        time.sleep(self.sleep_time_for_cpu_usage_check)
        container_statuses_end = self._get_all_stats_all_container()
        return self._calculate_usage_statistics(
            container_statuses_start, container_statuses_end
        )

    def get_stats_container(self, container_id):
        """
        Returns a list of dicts, with one element of the container with the id container_id the memory usage and
        the cpu utilization for all locally running docker containers
        """
        container_status_start = self._get_all_stats_container(container_id)
        time.sleep(self.sleep_time_for_cpu_usage_check)
        container_status_end = self._get_all_stats_container(container_id)
        return self._calculate_usage_statistics(
            [container_status_start], [container_status_end]
        )

    def get_information_all_containers(self):
        return self.containers

    def get_master_images(self, masterImage):
        # pull master image
        registry = os.getenv("HARBOR_URL").split("//")[-1]
        master_image_source = f"{registry}/{masterImage}"
        self.docker_client.images.pull(master_image_source, tag="latest")

    def _get_all_stats_all_container(self):
        """
        returns all the stats for all containers using the APIClient
        """
        container_statuses = []
        for container in self.containers:
            container_statuses.append(self.client.stats(container["Id"], stream=False))
        return container_statuses

    def _get_all_stats_container(self, container_id):
        return self.client.stats(container_id, stream=True)

    def _calculate_usage_statistics(self, start_measurements, end_measurements):
        stats = []
        for i in range(len(start_measurements)):
            cpu_usage = self._calculate_cpu_usage(
                start_measurements[i], end_measurements[i]
            )
            name_container = start_measurements[i]["name"]
            memory_usage = (
                start_measurements[i]["memory_stats"]["usage"]
                / start_measurements[i]["memory_stats"]["limit"]
            )
            status_dict = {
                "cpu_percent": cpu_usage,
                "memory_percent": memory_usage,
                "name": name_container,
            }
            stats.append(status_dict)
        return stats

    def _calculate_cpu_usage(self, start_measurement, end_measurement):
        cpu_cycles_container = (
            start_measurement["cpu_stats"]["cpu_usage"]["total_usage"]
            - end_measurement["cpu_stats"]["cpu_usage"]["total_usage"]
        )
        total_cycles = (
            start_measurement["cpu_stats"]["system_cpu_usage"]
            - end_measurement["cpu_stats"]["system_cpu_usage"]
        )
        return cpu_cycles_container / total_cycles


dockerClient = DockerClient()



import os
from typing import Union

from conductor_lib.src.torch import LightningTrainModel
from torch.utils.data import DataLoader

from station.common.clients.minio import MinioClient
from station.worker.loader import BaseLoader


class ModelTrainer:
    def __init__(
        self, model_id: str = None, data_set_id: str = None, station_url: str = None
    ):
        """
        Initialize a model trainer instance with a model and data set id

        :param model_id:
        :param data_set_id:
        :param station_url:
        """
        self.model_id = model_id
        self.station_url = station_url if station_url else os.getenv("STATION_API_URL")
        assert self.station_url
        self.model: LightningTrainModel = None
        # TODO fix minio
        self.data_set_id = data_set_id
        self.minio_client = MinioClient(
            minio_server="localhost:9000",
            secret_key="minio_admin",
            access_key="minio_admin",
        )

    def train_model(self, train_args: dict = None):
        pass

    def prepare_torch_data_loader(
        self,
        data_set_id: str = None,
        path: Union[str, os.PathLike] = None,
        data_config: dict = None,
    ) -> DataLoader:
        """
        Create a pytorch data loader based on a specified data set id or path, applies the transforms specified
        in the lightning model to the created data set

        :param data_set_id: identifier for a data set registered at the station
        :param path: path to directory accessible to the trainer where the raw data for the data set is stored
        :param data_config: dictionary containing additional configuration parameters for the data set
        :return:
        """
        if path:
            # TODO Load data from file system
            raise NotImplementedError("Only Minio based data sets available")
        else:
            # TODO use station api to get the relevant information for a data set from the db
            if data_set_id:
                self.data_set_id = data_set_id
            # create a loader that will handle loading and preparing the data from Minio
            base_loader = BaseLoader(
                data_set=data_set_id,
                station_api=self.station_url,
                config=data_config,
                transform=self.model.make_transform(),
            )
            data_loader = base_loader.make_data_loader()
            return data_loader


if __name__ == "__main__":
    model_id = "af8c8b1e-3d89-4500-bc7d-7888e13381bd"

    trainer = ModelTrainer(model_id=model_id, station_url="http://localhost:8001")
    db_model = trainer.get_model_from_station(model_id=model_id)
    print(db_model)



from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel


class TaskInstances(BaseModel):
    dag_id: str
    duration: Optional[float]
    end_date: Optional[datetime]
    execution_date: datetime
    executor_config: str
    hostname: str
    max_tries: int
    operator: str
    pid: Optional[int]
    priority_weight: int
    queue: str
    queued_when: Optional[datetime]
    start_date: Optional[datetime]
    state: Optional[str]
    task_id: str
    try_number: int
    unixname: str


class Tasklist(BaseModel):
    task_instances: List[TaskInstances]
    total_entries: int


class AirflowInformation(BaseModel):
    conf: dict
    dag_id: str
    dag_run_id: str
    end_date: Optional[datetime]
    execution_date: datetime
    external_trigger: bool
    logical_date: datetime
    start_date: Optional[datetime]
    state: str
    tasklist: Tasklist


class AirflowTaskLog(BaseModel):
    run_info: str


class AirflowRunMsg(BaseModel):
    train_id: str


class AirflowRun(BaseModel):
    run_id: str
    dag_id: str
    train_id: str
    start_date: datetime



from enum import Enum

import redis


class RedisJSONOps(str, Enum):
    SET = "JSON.SET"
    GET = "JSON.GET"


class Cache:
    def __init__(self, host="redis", port=6379, password=None, db=None):
        self.redis = redis.Redis(
            decode_responses=True, host=host, port=port, password=password, db=db
        )

    def set(self, key: str, value: str, ttl: int = 3600) -> None:
        """
        Set a key/value pair in the cache. With ttl.
        Args:
            key: key to set
            value: value of the key
            ttl: time until the key expires

        Returns:

        """
        self.redis.set(key, value, ex=ttl)

    def get(self, key: str) -> str:
        """
        Get a key from the cache.
        Args:
            key: cache key

        Returns:
            value of the key or None if not found
        """
        return self.redis.get(key)

    def json_set(self, key: str, value: str, ttl: int = 3600) -> None:
        """
        Adds a json string to the cache. With ttl.
        Args:
            key: key under which the json string is stored
            value: json string
            ttl: time until the key expires

        Returns:

        """
        self.redis.execute_command(
            RedisJSONOps.SET.value,
            key,
            ".",
            value,
        )
        self.redis.expire(key, ttl)

    def json_get(self, key) -> str:
        """
        Get a json string from the cache.
        Args:
            key: key under which the json string is stored

        Returns:
            json string or None if not found
        """
        json_string = self.redis.execute_command(RedisJSONOps.GET.value, key, ".")
        return json_string


redis_cache = None



import copy
import os.path
import sys
from typing import TYPE_CHECKING, Any, List

import click
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from station.common.clients.central.central_client import CentralApiClient
from station.common.config.fix import ConfigItemFix, GeneratorArg
from station.common.config.generators import GeneratorResult, generate_private_key
from station.common.config.station_config import StationConfig, StationSettings
from station.common.constants import Icons, PHTDirectories
from station.ctl.install.certs import generate_certificates

if TYPE_CHECKING:
    from station.ctl.config.validate import ValidationResult


def get_fixes_from_errors(config_dict: dict, errors: list[dict]) -> list[ConfigItemFix]:
    """Takes a list of validation errors resulting from initializing the station config pydantic model and
    returns a list of fixes that can be applied to the config to fix the errors.

    Args:
        errors: list of pydantic validation error dicts

    Returns:
        list of fixes that can be applied to the config to fix the errors
    """
    if len(errors) == 0:
        return []

    fixes = []

    # construct the config
    fixed_config = copy.deepcopy(config_dict)
    config = StationConfig.construct(**fixed_config)
    for error in errors:
        fix = get_fix_by_loc(
            config,
            error["loc"],
        )

        if fix:
            fixes.append(fix)
        else:
            unknown_fix = ConfigItemFix.no_fix(
                suggestion="Unable to fix error automatically or find suggestions. Please enter manually or "
                "contact the PHT team for help."
            )
            unknown_fix.issue = error["msg"]
            fixes.append(unknown_fix)
    return fixes


def get_fix_by_loc(config: StationConfig, loc: tuple) -> ConfigItemFix | None:
    """
    Returns the fix for the given location
    """

    level = len(loc)

    # get config attribute based on loc
    if level == 1:
        return config.get_fix(loc[0])
    # if it is a nested config item get the fix from the submodel
    else:
        sub_settings = _get_settings_for_loc(config, loc=loc)
        fix = sub_settings.get_fix(loc[-1])
        return fix


def _get_settings_for_loc(settings: StationSettings, loc: tuple) -> StationSettings:
    """
    Return the subsettings of the station settings for the given location
    """
    name = loc[0]

    if len(loc) == 1:
        return settings
    # get the model for the name
    model = getattr(settings, name)
    # get the sub model for the rest of the loc
    return _get_settings_for_loc(model, loc[1:])


def fix_config(ctx: dict, config: dict, results: List["ValidationResult"]) -> dict:
    """
    Allows for interactive fixes of issues in the station configuration
    Args:
        config: initial dictionary containing the config in the config yaml file
        results: validation results of the given dictionary

    Returns:
        updated config dictionary
    """
    fixed_config = copy.deepcopy(config)
    for result in results:
        # if the fix is a generator function, run it and set the value(s) in the config based on the results
        if result.fix.generator_function:
            prompt_text = "Manually enter a value or enter GENERATE to interactively generate a value"
            prompt_result = click.prompt(prompt_text, default="GENERATE")
            if prompt_result == "GENERATE":
                run_fix_generator(
                    fixed_config,
                    result.fix.generator_function,
                    result.fix.generator_args,
                )
            else:
                _set_config_value(fixed_config, ".".join(result.loc), prompt_result)

        else:
            fixed_value = click.prompt(
                result.fix.suggestion,
                default=result.fix.fix if result.fix.fix else None,
            )
            _set_config_value(fixed_config, ".".join(result.loc), fixed_value)

    return fixed_config


def run_fix_generator(
    config: dict, generator: callable, args: list[GeneratorArg] | None
) -> Any:
    """
    Runs a generator function and the configured arguments in an interactive prompt
    """

    if args is None:
        args = []

    arg_dict = {}
    # iterate over the arguments and prompt the user for the values
    for arg in args:
        value = click.prompt(arg.prompt, default="" if not arg.required else None)
        # prompt again if the value is required and not given
        if not value and arg.required:
            click.echo(
                f"{Icons.FAIL.value} {arg.arg} is a required argument for the generator function."
            )
            value = click.prompt(
                arg.prompt,
            )
            if not value:
                click.echo(
                    f"{Icons.FAIL.value} {arg.arg} is a required argument for the generator function. Exiting..."
                )
                sys.exit(1)
        arg_dict[arg.arg] = value

    # Get the result from the generator function
    result: list[GeneratorResult] | GeneratorResult = generator(**arg_dict)

    # Apply the result to the config
    if isinstance(result, list):
        for item in result:
            _set_config_value(config, ".".join(item.loc), item.value)
    else:
        _set_config_value(config, ".".join(result.loc), result.value)

    return


def _fix_certs(config: dict, strict: bool, install_dir: str):
    if strict:
        click.echo(
            "Valid certificates are required when not in development mode", err=True
        )
        sys.exit(1)

    domain = config["https"]["domain"]
    click.echo(f"Generating certificates for domain {domain}...")

    cert_dir = os.path.join(install_dir, PHTDirectories.CERTS_DIR.value)

    if not os.path.exists(cert_dir):
        os.makedirs(cert_dir)
    cert_path = os.path.join(cert_dir, "cert.pem")
    key_path = os.path.join(cert_dir, "key.pem")
    generate_certificates(domain, cert_path=str(cert_path), key_path=str(key_path))
    click.echo(f"Certificates generated at {cert_dir}")

    cert_list = config["https"].get("certs", [])
    if not cert_list:
        cert_list = []

    host_path = config.get("host_path", None)
    if host_path:
        cert_dir = os.path.join(host_path, PHTDirectories.CERTS_DIR.value)
        cert_path = os.path.join(cert_dir, "cert.pem")
        key_path = os.path.join(cert_dir, "key.pem")
    cert_paths = {"cert": str(cert_path), "key": str(key_path)}
    cert_list.append(cert_paths)
    config["https"]["certs"] = cert_list


def _fix_certs_path(config: dict, index: int, strict: bool = False):
    cert_path = config["https"]["certs"][index]["cert"]
    key_path = config["https"]["certs"][index]["key"]

    print(cert_path, key_path)
    if not cert_path and not key_path:
        _fix_certs(config, strict, config.get("install_dir", os.getcwd()))
        return
    if not os.path.isfile(cert_path):
        cert_path = click.prompt(
            f"Certificate at {cert_path} does not exist. "
            f"Please enter the correct path to the certificate file"
        )
        if not os.path.isfile(cert_path):
            click.echo("Certificate path is invalid", err=True)
        cert_path = str(os.path.abspath(cert_path))
    if not os.path.isfile(key_path):
        key_path = click.prompt(
            f"Key at {key_path} does not exist. "
            f"Please enter the correct path to the key file"
        )
        if not os.path.isfile(key_path):
            click.echo("Key path is invalid", err=True)
        key_path = str(os.path.abspath(key_path))

    config["https"]["certs"][index]["cert"] = cert_path
    config["https"]["certs"][index]["key"] = key_path


def _fix_private_key(config: dict) -> str:
    path = click.prompt(
        "Private key is missing enter the path to the private key "
        "file or press enter to generate a new one",
        default="GENERATE",
    )
    if path and path != "GENERATE":
        if not os.path.isfile(path):
            raise click.BadParameter(f"{path} is not a file")
        else:
            return path
    name = click.prompt("Name your private key file")
    passphrase = click.prompt(
        "Enter your passphrase. If given, it will be used to encrypt the private key",
        default="",
    )
    private_key_path, private_key, public_key = generate_private_key(
        name, config.get("install_dir"), passphrase
    )
    if passphrase:
        config["central"]["private_key_password"] = passphrase

    # if a host path is given append the name of the private key to this path
    host_path = config.get("host_path", None)
    if host_path:
        private_key_path = os.path.join(host_path, private_key_path)
        click.echo(
            f"Private key will be saved at: {private_key_path} on the host machine"
        )
    else:
        private_key_path = os.path.abspath(private_key_path)
        click.echo(f"New private key created: {private_key_path}")

    click.echo("Submitting public key to central API...", nl=False)
    # submit public key to central API
    _submit_public_key(config, public_key)
    click.echo(Icons.CHECKMARK.value)

    return private_key_path


def _set_config_value(config: dict, field: str, value: Any):
    nested_fields = field.split(".")
    if len(nested_fields) == 1:
        config[field] = value
    else:
        sub_config = None
        for field in nested_fields[:-1]:
            sub_config = config.get(field, {})
        sub_config[nested_fields[-1]] = value


def _submit_public_key(config: dict, public_key: rsa.RSAPublicKey):
    client = CentralApiClient(
        config["central"]["api_url"],
        robot_id=config["central"]["robot_id"],
        robot_secret=config["central"]["robot_secret"],
    )

    hex_key = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    ).hex()
    client.update_public_key(config["station_id"], hex_key)



from dotenv import find_dotenv, load_dotenv

from station.app.clients import StationClients
from station.app.settings import Settings

load_dotenv(find_dotenv())

settings = Settings()
# settings.setup()

clients = StationClients(settings)
# clients.initialize()

cache = None



from typing import List, Union

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from station.app.api import dependencies
from station.app.crud.crud_docker_trains import docker_trains
from station.app.crud.crud_train_configs import docker_train_config
from station.app.schemas.docker_trains import (
    DockerTrain,
    DockerTrainConfig,
    DockerTrainConfigCreate,
    DockerTrainConfigUpdate,
    DockerTrainCreate,
    DockerTrainExecution,
    DockerTrainSavedExecution,
    DockerTrainState,
)
from station.app.trains.docker import airflow

router = APIRouter()


@router.post("/sync", response_model=List[DockerTrain])
def synchronize_database(db: Session = Depends(dependencies.get_db)):
    return docker_trains.synchronize_central(db)


@router.get("", response_model=List[DockerTrain])
def get_available_trains(limit: int = 0, db: Session = Depends(dependencies.get_db)):
    if limit != 0:
        db_trains = docker_trains.get_multi(db, limit=limit)
    else:
        db_trains = docker_trains.get_multi(db)
    return db_trains


@router.post("", response_model=DockerTrain)
def register_train(
    create_msg: DockerTrainCreate, db: Session = Depends(dependencies.get_db)
):
    if docker_trains.get_by_train_id(db, train_id=create_msg.train_id):
        raise HTTPException(
            status_code=400,
            detail=f"Train with id '{create_msg.train_id}' already exists.",
        )
    db_train = docker_trains.create(db, obj_in=create_msg)
    return db_train


@router.get("/{train_id}", response_model=DockerTrain)
def get_train_by_train_id(
    train_id: Union[int, str], db: Session = Depends(dependencies.get_db)
):
    if isinstance(train_id, str):
        db_train = docker_trains.get_by_train_id(db, train_id)
    else:
        db_train = docker_trains.get(db, id=train_id)
    if not db_train:
        raise HTTPException(
            status_code=404, detail=f"Train with id '{train_id}' not found."
        )
    return db_train


@router.post("/{train_id}/run", response_model=DockerTrainSavedExecution)
def run_docker_train(
    train_id: str,
    run_config: DockerTrainExecution = None,
    db: Session = Depends(dependencies.get_db),
):
    execution = airflow.run_train(db, train_id, run_config)
    return execution


@router.get("/{train_id}/config", response_model=DockerTrainConfig)
def get_config_for_train(train_id: str, db: Session = Depends(dependencies.get_db)):
    train = docker_trains.get_by_train_id(db, train_id)
    if not train.config_id:
        raise HTTPException(
            status_code=404,
            detail=f"Train '{train_id}' does not have an assigned config.",
        )
    config = docker_train_config.get(db, train.config_id)
    return config


@router.post("/{train_id}/config/{config_id}", response_model=DockerTrain)
def assign_config_to_docker_train(
    train_id: str, config_id: int, db: Session = Depends(dependencies.get_db)
):
    train = docker_trains.get_by_train_id(db, train_id=train_id)
    if not train:
        raise HTTPException(
            status_code=404, detail=f"Train with id '{train_id}' not found."
        )

    config = docker_train_config.get(db, config_id)
    if not config:
        raise HTTPException(
            status_code=404, detail=f"Config with id '{config_id}' not found."
        )

    train = docker_train_config.assign_to_train(db, train_id, config.id)
    return train


@router.get("/{train_id}/state", response_model=DockerTrainState)
def get_state_for_train(train_id: str, db: Session = Depends(dependencies.get_db)):
    state = docker_trains.read_train_state(db, train_id)
    return state


@router.put("/{train_id}/state", response_model=DockerTrainState)
def update_state_for_train(
    train_id: str, state: DockerTrainState, db: Session = Depends(dependencies.get_db)
):
    state = docker_trains.update_train_state(db, train_id, state)
    return state


@router.get("/configs/all", response_model=List[DockerTrainConfig])
def get_all_docker_train_configs(
    db: Session = Depends(dependencies.get_db), skip: int = 0, limit: int = 100
):
    db_configs = docker_train_config.get_multi(db, skip=skip, limit=limit)
    return db_configs


@router.post("/config", response_model=DockerTrainConfig)
def add_docker_train_configuration(
    config_in: DockerTrainConfigCreate, db: Session = Depends(dependencies.get_db)
):
    print(config_in)
    if docker_train_config.get_by_name(db, name=config_in.name):
        raise HTTPException(
            status_code=400, detail="A config with the given name already exists."
        )
    config = docker_train_config.create(db, obj_in=config_in)
    return config


@router.put("/config/{config_id}", response_model=DockerTrainConfig)
def update_docker_train_configuration(
    update_config: DockerTrainConfigUpdate,
    config_id: int,
    db: Session = Depends(dependencies.get_db),
):
    old_config = docker_train_config.get(db, config_id)
    if not old_config:
        raise HTTPException(
            status_code=404, detail=f"Config with id '{config_id}' not found."
        )
    config = docker_train_config.update(db, db_obj=old_config, obj_in=update_config)
    return config


@router.get("/config/{config_id}", response_model=DockerTrainConfig)
def get_docker_train_configuration(
    config_id: int, db: Session = Depends(dependencies.get_db)
):
    config = docker_train_config.get(db, config_id)
    if not config:
        raise HTTPException(
            status_code=404, detail=f"Config with id '{config_id}' not found."
        )

    return config


@router.get("/executions/all", response_model=List[DockerTrainSavedExecution])
def get_all_docker_train_executions(
    skip: int = 0, limit: int = 100, db: Session = Depends(dependencies.get_db)
):
    db_executions = docker_trains.get_executions(db, skip=skip, limit=limit)
    return db_executions


@router.get("/{train_id}/executions", response_model=List[DockerTrainSavedExecution])
def get_docker_train_executions(
    train_id: str,
    skip: int = 0,
    limit: int = 100,
    db: Session = Depends(dependencies.get_db),
):
    executions = docker_trains.get_train_executions(
        db,
        train_id,
    )
    return executions



# @pytest.fixture
# def auth_values():
#     load_dotenv(find_dotenv())
#     url = f"http://{os.getenv('AUTH_SERVER_HOST')}:{os.getenv('AUTH_SERVER_PORT')}/token"
#     robot_id = os.getenv('AUTH_ROBOT_ID')
#     robot_secret = os.getenv('AUTH_ROBOT_SECRET')
#     return url, robot_id, robot_secret
#
#
# def test_get_robot_token(auth_values):
#     url, robot_id, robot_secret = auth_values
#     if os.getenv("ENVIRONMENT") in ["development", "testing"]:
#         token = get_robot_token(token_url=url, robot_id=robot_id, robot_secret=robot_secret)
#
#         assert token
#
#
# def test_validate_user_token(auth_values):
#     if os.getenv("ENVIRONMENT") in ["development", "testing"]:
#         url, robot_id, robot_secret = auth_values
#         token = get_robot_token(token_url=url, robot_id=robot_id, robot_secret=robot_secret)
#
#         # get user token
#         data = {"username": "admin", "password": "start123"}
#         response = requests.post(url, json=data)
#         user_token = response.json()["access_token"]
#
#         response = validate_user_token(token_url=url, token=user_token, robot_token=token)
#         assert response.name == "admin"
#



import os

import requests
from requests.auth import HTTPBasicAuth
from train_lib.clients.fhir import build_query_string
from train_lib.clients.fhir.fhir_client import BearerAuth


class FhirClient:
    def __init__(
        self,
        server_url: str = None,
        username: str = None,
        password: str = None,
        token: str = None,
        server_type: str = None,
        disable_auth: bool = False,
    ):
        """
        Fhir client for station internal / health check / how many data points are in a fhir server
        The code in hear is mostly copied from the train libary fhir client .

        :param server_url: base url of the FHIR server to execute the query against
        :param username: username for use in basic auth authentication against the fhir server
        :param password: password for use in basic auth
        :param token: token to use for authenticating against a FHIR server using a bearer token
        :param server_type: the type of the server one of ["blaze", "hapi", "ibm"]
        """
        self.server_url = server_url if server_url else os.getenv("FHIR_ADDRESS")
        self.username = username if username else os.getenv("FHIR_USER")
        self.password = password if password else os.getenv("FHIR_PW")
        self.token = token if token else os.getenv("FHIR_TOKEN")
        self.server_type = server_type if server_type else os.getenv("FHIR_SERVER_TYPE")
        self.output_format = None
        # Check for correct initialization based on env vars or constructor parameters
        if not (self.username and self.password) and self.token:
            raise ValueError("Only one of username:pw or token auth can be selected")
        if not ((self.username and self.password) or self.token) and disable_auth:
            raise ValueError(
                "Insufficient login information, either token or username and password need to be set."
            )
        if not self.server_url:
            raise ValueError("No FHIR server address available")

    def health_check(self):
        api_url = self._generate_api_url() + "/metadata"
        auth = self._generate_auth()

        response = {"status": None, "date": None, "name": None}
        try:
            # TODO remove verify false only for thesting becouse of verificaion problems with the ibm fhir server
            r = requests.get(api_url, auth=auth, verify=False)
            r.raise_for_status()
            r_json = r.json()

            # This is hear to make the response the same for all health check
            if r_json["status"] == "active":
                response["status"] = "healthy"
                response["date"] = r_json["date"]
        except requests.exceptions.ConnectionError as e:
            print(e)
            pass
        except requests.exceptions.HTTPError as e:
            print(e)
            pass
        return response

    def get_number_of_resource(self):
        api_url = self._generate_api_url() + "/Resource?_count=0"
        auth = self._generate_auth()
        r = requests.get(api_url, auth=auth, verify=False).json()
        return r["total"]

    def _generate_url(
        self,
        query: dict = None,
        query_string: str = None,
        return_format="json",
        limit=1000,
    ):
        """
        Generates the fhir search url to request from the server based. Either based on a previously given query string
        or based on a dictionary containing the query definition in the query.json file.

        :param query: dictionary containing the content of the query definition in the query.json file
        :param query_string: Predefined Fhir url query string to append to base server url
        :param return_format: the format in which the server should return the data.
        :param limit: the max number of entries in a paginated response
        :return: url string to perform a request against a fhir server with
        """
        url = self._generate_api_url() + "/"
        if query:
            url += build_query_string(query_dict=query)
        elif query_string:
            url += query_string
        else:
            raise ValueError("Either query dictionary or string need to be given")
        # add formatting configuration
        url += f"&_format={return_format}&_limit={limit}"

        return url

    def _generate_auth(self) -> requests.auth.AuthBase:
        """
        Generate authoriation for the request to be sent to server. Either based on a given bearer token or using basic
        auth with username and password.

        :return: Auth object to pass to a requests call.
        """
        if self.username and self.password:
            return HTTPBasicAuth(username=self.username, password=self.password)
        # TODO request token from id provider if configured
        elif self.token:
            return BearerAuth(token=self.token)

    def _generate_api_url(self) -> str:
        url = self.server_url
        if self.server_type == "ibm":
            url += "/fhir-server/api/v4"

        elif self.server_type in ["blaze", "hapi"]:
            url += "/fhir"

        else:
            raise ValueError(f"Unsupported FHIR server type: {self.server_type}")

        return url



import click

import docker
import docker.errors
from docker import DockerClient
from station.common import constants


def setup_docker():
    client = _get_docker_client()
    click.echo("Creating docker volumes... ", nl=False)
    for vol in constants.DockerVolumes:
        try:
            volume = client.volumes.get(vol.value)
            if (
                click.prompt(
                    f"\nVolume {vol.value} already exists. Do you want to reset it? [y/N]",
                    default="N",
                )
                == "y"
            ):
                volume.remove()
                client.volumes.create(name=vol.value)
                click.echo(f"Volume was reset {vol.value}")

        except docker.errors.NotFound:
            try:
                client.volumes.create(name=vol.value)
                click.echo(constants.Icons.CHECKMARK.value)

            except Exception as e:
                click.echo(
                    f"{constants.Icons.CROSS.value} Failed to create volume: \n {e}"
                )
                raise e

        except Exception as e:
            click.echo(f"{constants.Icons.CROSS.value} Failed to create volume: \n {e}")
            raise e

    click.echo("Creating docker networks... ", nl=False)
    for net in constants.DockerNetworks:
        try:
            client.networks.get(net.value)
        except docker.errors.NotFound:
            try:
                client.networks.create(name=net.value)

            except Exception as e:
                click.echo(
                    f"{constants.Icons.CROSS.value} Failed to create network: \n {e}"
                )
                raise e
        except Exception as e:
            click.echo(
                f"{constants.Icons.CROSS.value} Failed to create network: \n {e}"
            )
            raise e
    click.echo(constants.Icons.CHECKMARK.value)


def download_docker_images(ctx):
    client = _get_docker_client()

    click.echo("Downloading PHT Station images:")
    version = ctx.obj["version"]
    # download pht images
    for image in constants.PHTImages:
        _download_image(client, image.value, version)
    click.echo("Downloading service images:")
    for image in constants.ServiceImages:
        _download_image(client, image.value)


def _download_image(client, image: str, tag=None):
    if not tag:
        image_tag = image.split(":")
        if len(image_tag) == 2:
            image = image_tag[0]
            tag = image_tag[1]

    click.echo(f"\tDownloading {image}:{tag}... ", nl=False)
    try:
        client.images.pull(image, tag=tag)
        click.echo(constants.Icons.CHECKMARK.value)

    except Exception as e:
        click.echo(f"{constants.Icons.CROSS.value} Failed to download image: \n {e}")


def _get_docker_client() -> DockerClient:
    try:
        client = docker.from_env()
    except Exception as e:
        click.echo(f"{constants.Icons.CROSS.value} Failed to connect to docker: \n {e}")
        raise e

    return client



from io import BytesIO
from typing import Any, List

from fastapi import APIRouter, Depends, File, HTTPException, UploadFile
from fastapi.responses import StreamingResponse
from loguru import logger
from sqlalchemy.orm import Session

from station.app.api import dependencies
from station.app.config import clients
from station.app.crud import datasets
from station.app.datasets import statistics
from station.app.schemas.datasets import (
    DataSet,
    DataSetCreate,
    DataSetStatistics,
    DataSetUpdate,
    MinioFile,
)
from station.common.constants import DataDirectories

router = APIRouter()


@router.post("", response_model=DataSet)
def create_new_data_set(
    create_msg: DataSetCreate,
    db: Session = Depends(dependencies.get_db),
) -> DataSet:
    dataset = datasets.get_by_name(db, name=create_msg.name)
    if dataset:
        raise HTTPException(
            status_code=400,
            detail=f"Dataset with name {create_msg.name} already exists.",
        )
    try:
        db_dataset = datasets.create(db, obj_in=create_msg)
        if not db_dataset:
            raise HTTPException(
                status_code=404, detail="Error while creating new dataset."
            )
        return db_dataset
    except FileNotFoundError:
        raise HTTPException(
            status_code=404,
            detail=f"Dataset file not found at {create_msg.access_path}.",
        )
    except NotImplementedError:
        raise HTTPException(
            status_code=422,
            detail=f"Storage type {create_msg.storage_type} not possible yet.",
        )


@router.get("", response_model=List[DataSet])
def read_all_data_sets(db: Session = Depends(dependencies.get_db)) -> List[DataSet]:
    all_datasets = datasets.get_multi(db=db, limit=None)
    return all_datasets


@router.get("/{data_set_id}", response_model=DataSet)
def get_data_set(
    data_set_id: Any, db: Session = Depends(dependencies.get_db)
) -> DataSet:
    db_dataset = datasets.get(db, data_set_id)
    if not db_dataset:
        raise HTTPException(status_code=404, detail="Dataset not found.")
    return db_dataset


@router.put("/{dataset_id}", response_model=DataSet)
def update_data_set(
    dataset_id: Any,
    update_msg: DataSetUpdate,
    db: Session = Depends(dependencies.get_db),
) -> DataSet:
    db_dataset = datasets.get(db, id=dataset_id)
    if not db_dataset:
        raise HTTPException(status_code=404, detail="Dataset not found.")
    new_db_dataset = datasets.update(db=db, db_obj=db_dataset, obj_in=update_msg)
    return new_db_dataset


@router.delete("/{dataset_id}", response_model=DataSet)
def delete_data_set(
    dataset_id: Any, db: Session = Depends(dependencies.get_db)
) -> DataSet:
    db_dataset = datasets.get(db, dataset_id)
    if not db_dataset:
        raise HTTPException(status_code=404, detail="Dataset not found.")
    db_data_set = datasets.remove(db, id=dataset_id)
    return db_data_set


@router.post("/{dataset_id}/files")
async def upload_data_set_file(
    dataset_id: str,
    files: List[UploadFile] = File(description="Multiple files as UploadFile"),
    db: Session = Depends(dependencies.get_db),
):
    db_dataset = datasets.get(db, dataset_id)
    if not db_dataset:
        raise HTTPException(status_code=404, detail=f"Dataset {dataset_id} not found.")
    if not files:
        raise HTTPException(status_code=400, detail="No files provided.")
    for file in files:
        if not file.filename:
            raise HTTPException(status_code=400, detail="No filename provided.")

    # minio_client = MinioClient()
    await clients.minio.save_dataset_files(db_dataset.id, files)


@router.get("/{data_set_id}/files", response_model=List[MinioFile])
async def get_data_set_files(
    data_set_id: str, file_name: str = None, db: Session = Depends(dependencies.get_db)
):
    db_dataset = datasets.get(db, data_set_id)
    if not db_dataset:
        raise HTTPException(status_code=404, detail=f"Dataset {data_set_id} not found.")

    items = clients.minio.get_minio_dir_items(DataDirectories.DATASETS, data_set_id)
    if file_name:
        pass
    return items


@router.delete("/{data_set_id}/files")
async def delete_file_from_dataset(
    data_set_id: str, file_name: str, db: Session = Depends(dependencies.get_db)
):
    db_dataset = datasets.get(db, data_set_id)
    if not db_dataset:
        raise HTTPException(status_code=404, detail=f"Dataset {data_set_id} not found.")

    clients.minio.delete_file(DataDirectories.DATASETS.value, file_name)


@router.get("/{data_set_id}/download", response_class=StreamingResponse)
async def download(
    data_set_id: Any,
    archive_type: str = "tar",
    db: Session = Depends(dependencies.get_db),
):
    db_dataset = datasets.get(db, data_set_id)
    if not db_dataset:
        raise HTTPException(status_code=404, detail="Dataset not found.")
    items = clients.minio.get_minio_dir_items(DataDirectories.DATASETS, data_set_id)

    if len(items) == 0:
        raise HTTPException(status_code=404, detail="No files found.")
    elif len(items) == 1:
        data = clients.minio.get_file(
            str(DataDirectories.DATASETS.value), items[0].full_path
        )
        obj = BytesIO(data)
        return StreamingResponse(content=obj, media_type="application/octet-stream")

    else:
        archive = clients.minio.make_dataset_archive(
            data_set_id, items=items, archive_type=archive_type
        )
        return StreamingResponse(content=archive, media_type="application/zip")


@router.get("/{data_set_id}/stats", response_model=DataSetStatistics)
def get_data_set_statistics(
    data_set_id: Any, file_name: str = None, db: Session = Depends(dependencies.get_db)
):
    db_dataset = datasets.get(db, data_set_id)
    if not db_dataset:
        raise HTTPException(status_code=404, detail="Dataset not found.")

    items = clients.minio.get_minio_dir_items(DataDirectories.DATASETS, data_set_id)
    if len(items) == 0 and not db_dataset.fhir_server:
        raise HTTPException(status_code=404, detail="No files found in the dataset.")

    if db_dataset.data_type in {"hybrid", "structured"}:
        if file_name:
            items = [item for item in items if item.file_name == file_name]
            if len(items) == 0:
                raise HTTPException(
                    status_code=404, detail=f"File {file_name} not found."
                )

        try:
            stats = statistics.load_stats(db_dataset.summary, file_name)
            logger.info(f"Loaded stats for {file_name}")
            return stats
        except ValueError as e:
            logger.info(
                f"Could not load existing stats for {db_dataset.id} {file_name}. {e}"
            )
        try:
            logger.info(f"Calculating stats for {db_dataset.id} {file_name}")
            file_content = clients.minio.get_file(
                DataDirectories.DATASETS, items[0].full_path
            )
            df = statistics.load_tabular(items[0], file_content)
            stats = statistics.get_dataset_statistics(df)
            datasets.add_stats(db, data_set_id, stats, items[0].file_name)
            return stats

        except TypeError:
            raise HTTPException(
                status_code=400,
                detail=f"File {file_name} is not in a supported tabular format.",
            )

    else:
        raise HTTPException(
            status_code=400,
            detail=f"Data type {db_dataset.data_type} is not supported yet.",
        )



import io
import json
from typing import Optional, Union

import pandas as pd
import plotly.express as px
import plotly.io
from loguru import logger
from pandas.api.types import is_bool_dtype, is_numeric_dtype
from plotly.graph_objects import Figure

from station.app.schemas.datasets import DataSetFigure, DataSetStatistics, MinioFile


def get_dataset_statistics(dataframe: pd.DataFrame) -> Optional[DataSetStatistics]:
    """
    Computes statistical information of a dataset
    :param dataframe: Dataset as dataframe object
    :return: Dataset statistics
    """
    if not (isinstance(dataframe, pd.DataFrame)):
        raise TypeError
    shape = dataframe.shape
    description = dataframe.describe(include="all")

    n_items = shape[0]
    n_features = shape[1]
    columns_inf = get_column_information(dataframe, description)

    schema_data = {
        "n_items": n_items,
        "n_features": n_features,
        "column_information": columns_inf,
    }

    statistics = DataSetStatistics(**schema_data)
    return statistics


def get_column_information(dataframe: pd.DataFrame, description: pd.DataFrame) -> dict:
    """
    Extract information out of dataframe and summarize it in a dictionary
    :param dataframe: Dataframe to summarize
    :param description: Dataframe description as dataframe object
    :return: Dictionary with column information
    """
    columns_inf = []
    columns = dataframe.columns.values.tolist()
    for i in range(len(columns)):
        title = columns[i]
        count = description[title]["count"]
        columns_inf.append({"title": title})
        nan_count = dataframe[title].isna().sum()
        if not (is_numeric_dtype(dataframe[title])) or is_bool_dtype(dataframe[title]):
            # extract information from categorical column
            columns_inf[i]["not_na_elements"] = count - nan_count
            columns_inf, chart_json = process_categorical_column(
                dataframe, columns_inf, i, description, title
            )
        else:
            # extract information from numerical column
            zero_count = dataframe[title][dataframe[title] == 0].count()
            undefined_count = nan_count + zero_count
            columns_inf[i]["not_na_elements"] = count - undefined_count
            columns_inf, chart_json = process_numerical_column(
                dataframe, columns_inf, i, description, title
            )

        if chart_json is not None:
            columns_inf[i]["figure"] = chart_json

    return columns_inf


def process_numerical_column(
    dataframe: pd.DataFrame,
    columns_inf: dict,
    i: int,
    description: pd.DataFrame,
    title: str,
) -> tuple[dict, DataSetFigure]:
    """
    Extract information from numerical column and create plot of column data
    :param dataframe: dataset as dataframe object
    :param columns_inf: array containing all information of different columns
    :param i: current column index
    :param description: description of all dataset columns
    :param title: title of column with index i
    :return: array with column information, key and json to save chart in cache
    """
    columns_inf[i]["type"] = "numeric"
    columns_inf[i]["mean"] = description[title]["mean"]
    columns_inf[i]["std"] = description[title]["std"]
    columns_inf[i]["min"] = description[title]["min"]
    columns_inf[i]["max"] = description[title]["max"]
    # create boxplot for numerical data
    fig = px.box(dataframe, y=title)
    chart_json = create_figure(fig)
    return columns_inf, chart_json


def process_categorical_column(
    dataframe: pd.DataFrame,
    columns_inf: dict,
    i: int,
    description: pd.DataFrame,
    title: str,
) -> tuple[dict, DataSetFigure]:
    """
    Extract information from categorical column and create plot of column data
    :param dataframe: dataset as dataframe object
    :param columns_inf: array containing all information of different columns
    :param i: current column index
    :param description: description of all dataset columns
    :param title: title of column with index i
    :return: array with column information, key and json to save chart in cache
    """
    count = columns_inf[i]["not_na_elements"]
    unique = description[title]["unique"]
    top = description[title]["top"]
    freq = description[title]["freq"]
    chart_json = None

    # if every entry has an unique value (or at most 50 values are given multiple times)
    if count - 50 < unique <= count:
        column_type = "unique"
        columns_inf[i]["type"] = column_type
        if unique != count:
            columns_inf[i]["number_of_duplicates"] = count - unique
    else:
        # all elements of column have the same value
        if unique == 1:
            column_type = "equal"
            columns_inf[i]["value"] = top
            # if column has just one equal value, no plot is created

        else:
            column_type = "categorical"
            columns_inf[i]["number_categories"] = unique
            columns_inf[i]["most_frequent_element"] = top
            columns_inf[i]["frequency"] = freq

            # pie chart if number of classes is below 6
            # value counts are provided if there are less then 6 classes
            if unique < 6:
                columns_inf[i]["value_counts"] = dict(dataframe[title].value_counts())
                fig = px.pie(dataframe, names=title, title=title)

            # histogram if number of classes is greater than 10
            elif unique >= 6:
                fig = px.histogram(dataframe, x=title, title=title)
                fig.update_layout(yaxis_title="Anzahl")

            if fig is not None:
                chart_json = create_figure(fig)

        columns_inf[i]["type"] = column_type

    return columns_inf, chart_json


def create_figure(fig: Figure) -> DataSetFigure:
    """
    Create DataSetFigure-Object of a plotly figure
    :param fig: Plotly figure
    :return: DataSetFigure-Object with JSON-representation of plotly figure
    """
    fig_json = plotly.io.to_json(fig)
    obj = json.loads(fig_json)
    figure = DataSetFigure(fig_data=obj)
    return figure


def load_tabular(file: MinioFile, content: bytes) -> pd.DataFrame:
    """
    Load tabular data from file
    :param file: File to load
    :param content: Content of file
    :return: Dataframe with data from file
    """

    extension = file.file_name.split(".")[-1].lower()
    print("extension: ", extension)
    if extension == "csv":
        dataframe = pd.read_csv(io.BytesIO(content))
    elif extension == "xlsx":
        dataframe = pd.read_excel(io.BytesIO(content))
    elif extension == "json":
        dataframe = pd.read_json(io.BytesIO(content))
    else:
        raise TypeError

    return dataframe


def load_stats(
    stats_json: Union[str, dict] = None, file_name: str = None
) -> DataSetStatistics:
    """
    Load statistics from json string
    :param stats_json: json string containing statistics
    :param file_name: name of file containing statistics
    :return: DataSetStatistics object
    """

    if stats_json is None:
        raise ValueError("No statistics available")

    if isinstance(stats_json, str):
        stats_json = json.loads(stats_json)
        return DataSetStatistics(**stats_json)
    else:
        if file_name:
            file_stats = stats_json.get(file_name)
            if file_stats is None:
                logger.warning(f"Statistics for file {file_name} not found")
                raise ValueError("No statistics available")
            else:
                stats = DataSetStatistics(**file_stats)
                return stats
        else:
            stats = DataSetStatistics(**stats_json)
            return stats



import pathlib

import uvicorn
from dotenv import find_dotenv, load_dotenv

from station.app.cache import Cache
from station.app.config import clients, settings
from station.app.db.setup_db import setup_db
from station.app.settings import StationRuntimeEnvironment


def setup():
    load_dotenv(find_dotenv())
    # settings.setup()

    # minio
    # minio_client = MinioClient()
    # minio_client.setup_buckets()


def main():
    load_dotenv(find_dotenv())
    # setup()
    settings.setup()

    setup_db(dev=False, reset=False)
    clients.initialize()
    # cache.redis_cache = cache.Cache(settings.config.redis.host)
    Cache(
        host=settings.config.redis.host,
        port=settings.config.redis.port,
        db=settings.config.redis.db,
        password=settings.config.redis.password,
    )
    # # Configure logging behaviour
    # log_config = uvicorn.config.LOGGING_CONFIG
    # log_config["formatters"]["access"][
    #     "fmt"
    # ] = "%(asctime)s - %(levelname)s - %(message)s"
    # log_config["formatters"]["default"][
    #     "fmt"
    # ] = "%(asctime)s - %(levelname)s - %(message)s"

    project_root = pathlib.Path(__file__).parent.parent
    uvicorn.run(
        "station.app.main:app",
        port=settings.config.port,
        host=settings.config.host,
        reload=settings.config.environment == StationRuntimeEnvironment.DEVELOPMENT,
        reload_dirs=[str(project_root)],
        # log_config=log_config,
    )


if __name__ == "__main__":
    main()



from loguru import logger

from station.app.settings import Settings
from station.common.clients.airflow.client import AirflowClient
from station.common.clients.central.central_client import CentralApiClient
from station.common.clients.harbor_client import HarborClient
from station.common.clients.minio.client import MinioClient


class StationClients:
    _airflow: AirflowClient
    _harbor: HarborClient
    _minio: MinioClient
    _central: CentralApiClient

    def __init__(self, settings: Settings):
        self.settings = settings
        self.is_initialized = False

    def initialize(self):
        logger.info("Initializing clients")
        if not self.settings.is_initialized:
            logger.warning(
                "Station settings are not initialized. Please call settings.setup() before initializing clients."
            )
            self.settings.setup()
        self._airflow = AirflowClient(
            airflow_api_url=self.settings.config.airflow.api_url,
            airflow_user=self.settings.config.airflow.user,
            airflow_password=self.settings.config.airflow.password.get_secret_value(),
        )

        # check for secretstr and get secret value if needed
        if isinstance(self.settings.config.registry.password, str):
            registry_password = self.settings.config.registry.password
        else:
            registry_password = (
                self.settings.config.registry.password.get_secret_value()
            )

        self._harbor = HarborClient(
            api_url=self.settings.config.registry.api_url,
            username=self.settings.config.registry.user,
            password=registry_password,
        )

        self._minio = MinioClient(
            minio_server=self.settings.config.minio.server_url,
            access_key=self.settings.config.minio.access_key,
            secret_key=self.settings.config.minio.secret_key,
        )

        self._minio.setup_buckets()

        if isinstance(self.settings.config.central_ui.robot_secret, str):
            robot_secret = self.settings.config.central_ui.robot_secret
        else:
            robot_secret = (
                self.settings.config.central_ui.robot_secret.get_secret_value()
            )

        self._central = CentralApiClient(
            base_url=self.settings.config.central_ui.api_url,
            robot_id=self.settings.config.central_ui.robot_id,
            robot_secret=robot_secret,
        )

        self.is_initialized = True
        logger.info("Clients initialized")

    @property
    def airflow(self) -> AirflowClient:
        if not self.is_initialized:
            logger.warning(
                "Station clients are not initialized. Please call clients.initialize() before using clients."
            )
            self.initialize()
        return self._airflow

    @property
    def harbor(self) -> HarborClient:
        if not self.is_initialized:
            logger.warning(
                "Station clients are not initialized. Please call clients.initialize() before using clients."
            )
            self.initialize()
        return self._harbor

    @property
    def minio(self) -> MinioClient:
        if not self.is_initialized:
            logger.warning(
                "Station clients are not initialized. Please call clients.initialize() before using clients."
            )
            self.initialize()
        return self._minio

    @property
    def central(self) -> CentralApiClient:
        if not self.is_initialized:
            logger.warning(
                "Station clients are not initialized. Please call clients.initialize() before using clients."
            )
            self.initialize()
        return self._central



import os
from typing import Union

import pytorch_lightning as pl
from torch.utils.data import DataLoader, Dataset
from torchvision.transforms import Compose, ToTensor

from station.common.clients.minio import MinioClient
from station.worker.loader.dataset import MinioFolderDS


class BaseLoader:
    def __init__(
        self,
        data_set: Union[Dataset, str] = None,
        station_api: str = None,
        config: dict = None,
        transform: Compose = None,
    ):
        self.station_api = station_api if station_api else os.getenv("STATION_API_URL")
        self.data_set = data_set
        self.transform = transform
        self.config = config
        # TODO fix minio
        self.minio_client = MinioClient()

    def make_data_loader(self) -> DataLoader:
        """
        Create a pytorch data loder based on this instance's fields, either takes a data set directly passed
        to the constructor or loads the data set from Minio if the given data set is specified as a string identifier

        :return: pytorch Dataloader to be used in model training
        """
        if isinstance(self.data_set, str):
            # Get data from db
            self.data_set = self.get_data_set()
        if self.config:
            return DataLoader(self.data_set, **self.config)
        return DataLoader(self.data_set)

    def make_lightning_data_module(self) -> pl.LightningDataModule:
        pass

    def get_data_set(self):
        # TODO enable different data set types
        dataset = MinioFolderDS(
            client=self.minio_client,
            data_set_id=self.data_set,
            transform=self.transform,
        )
        return dataset


if __name__ == "__main__":
    transform = Compose([ToTensor()])

    loader = BaseLoader(
        data_set="cifar/batch_1",
        transform=transform,
        config={"batch_size": 6, "num_workers": 0},
    ).make_data_loader()
    for i, batch in enumerate(loader):
        print(batch.size())
        if i > 3:
            break




import pytest
from fastapi.testclient import TestClient
from requests import HTTPError

from station.app.api.dependencies import get_db
from station.app.main import app

from .test_db import override_get_db

app.dependency_overrides[get_db] = override_get_db

client = TestClient(app)


@pytest.fixture
def train_id():
    response = client.post(
        "/api/local-trains",
        json={
            "name": "test_train",
            "custom_image": "test_image",
        },
    )
    return response.json()["id"]


def test_create_local_train():
    response = client.post(
        "/api/local-trains",
        json={
            "name": "test_train",
            "custom_image": "test_image",
        },
    )
    assert response.status_code == 200
    assert response.json()["name"] == "test_train"


def test_get_local_train(train_id):
    response = client.get(f"/api/local-trains/{train_id}")
    assert response.status_code == 200
    assert response.json()["name"] == "test_train"


def test_update_local_train(train_id):
    response = client.put(
        f"/api/local-trains/{train_id}",
        json={
            "name": "test_train_updated",
        },
    )

    assert response.status_code == 200
    assert response.json()["name"] == "test_train_updated"


def test_delete_local_train(train_id):
    response = client.delete(f"/api/local-trains/{train_id}")
    assert response.status_code == 200
    assert response.json()["name"] == "test_train"

    r = client.get(f"/api/local-trains/{train_id}")
    with pytest.raises(HTTPError):
        r.raise_for_status()



import json
import os
import sys
import time
from typing import Tuple

import click
from rich.console import Console

import docker
from station.common.clients.central.central_client import CentralApiClient
from station.common.constants import Icons, PHTDirectories, PHTImages
from station.ctl.config import fix_config, validate_config
from station.ctl.config.command import render_config
from station.ctl.install import templates
from station.ctl.install.docker import setup_docker
from station.ctl.install.fs import check_create_pht_dirs


@click.command(help="Install the station software based on the configuration file.")
@click.option(
    "--install-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Install location for station software. Defaults to current working directory.",
)
@click.option(
    "--host-path",
    type=str,
    help="Host path for containerized execution of the installer",
    required=False,
)
@click.pass_context
def install(ctx, install_dir, host_path):
    # validate configuration before installing
    click.echo("Validating configuration... ", nl=False)
    ctx.obj["host_path"] = host_path
    if not install_dir:
        install_dir = os.getcwd()
    ctx.obj["install_dir"] = install_dir
    result = validate_config(ctx.obj["station_config"], host_path=host_path)

    if result is not None:
        table, issues = result
        click.echo(Icons.CROSS.value)
        console = Console()
        console.print(table)

        if ctx.obj["station_config"]["environment"] == "production":
            click.confirm(
                "Station configuration is invalid. Please fix the errors displayed above. \n"
                "Would you like to fix the configuration now?",
                abort=True,
            )
            fixed_config = fix_config(ctx.obj, ctx.obj["station_config"], issues)
            # merge config with fixed config
            ctx.obj["station_config"] = ctx.obj["station_config"] | fixed_config
            render_config(ctx.obj["station_config"], ctx.obj["config_path"])

        else:
            res = click.prompt(
                "Station configuration is invalid. Press enter to fix the configuration now. "
                "[dev mode detected enter (skip) to continue]",
                default="y",
            )
            if res == "y":
                fixed_config = fix_config(ctx.obj, ctx.obj["station_config"], issues)
                ctx.obj["station_config"] = ctx.obj["station_config"] | fixed_config
                render_config(ctx.obj["station_config"], ctx.obj["config_path"])
            else:
                click.echo("Skipping configuration fix.")

    else:
        click.echo(Icons.CHECKMARK.value)

    click.echo(
        "Installing station software to {}".format(
            host_path if host_path else install_dir
        )
    )
    # ensure file system is set up
    check_create_pht_dirs(install_dir)

    # get credentials for registry
    reg_credentials = _request_registry_credentials(ctx)
    ctx.obj["registry"] = reg_credentials.dict()

    # setup docker volumes and networks
    setup_docker()
    # download docker images
    # download_docker_images(ctx)

    # setup_auth_server
    # _setup_auth_server(ctx)

    # copy certificates to install dir
    # copy_certificates(ctx)

    # render templates according to configuration and store output paths in configuration object
    ctx.obj["init_sql_path"] = write_init_sql(ctx)
    traefik_config_path, router_config_path = write_traefik_configs(ctx)
    ctx.obj["traefik_config_path"] = traefik_config_path
    ctx.obj["router_config_path"] = router_config_path
    ctx.obj["airflow_config_path"] = write_airflow_config(ctx)
    ctx.obj["authup_config_path"] = write_authup_config(ctx)

    # render the updated configuration file
    render_config(ctx.obj, ctx.obj["config_path"])

    # render the final compose template
    write_compose_file(ctx)


def _request_registry_credentials(ctx):
    click.echo("Requesting registry credentials from central api... ", nl=False)
    url = ctx.obj["station_config"]["central"]["url"]
    client = ctx.obj["station_config"]["central"]["robot_id"]
    secret = ctx.obj["station_config"]["central"]["robot_secret"]
    client = CentralApiClient(url, client, secret)

    credentials = client.get_registry_credentials(ctx.obj["station_config"]["id"])
    click.echo(Icons.CHECKMARK.value)

    return credentials


def _setup_auth_server(ctx):
    click.echo("Setting up auth server... ", nl=False)
    client = docker.from_env()

    auth_image = f"{PHTImages.AUTH.value}:{ctx.obj['version']}"
    command = "setup"

    if ctx.obj.get("host_path"):
        writable_dir = os.path.join(
            ctx.obj["host_path"], str(PHTDirectories.SERVICE_DATA_DIR.value), "auth"
        )
    else:
        writable_dir = os.path.join(
            ctx.obj["install_dir"], str(PHTDirectories.SERVICE_DATA_DIR.value), "auth"
        )

    auth_volumes = {str(writable_dir): {"bind": "/usr/src/app/writable", "mode": "rw"}}
    environment = {
        "ADMIN_USER": "admin",
        "ADMIN_PASSWORD": ctx.obj["admin_password"],
        "NODE_ENV": "production",
        "PUBLIC_URL": "https://" + ctx.obj["https"]["domain"] + "/auth",
        "AUTHORIZATION_REDIRECT_URL": "https://" + ctx.obj["https"]["domain"],
        "ROBOT_ENABLED": "true",
        # "TYPEORM_CONNECTION": "postgres",
        # "TYPEORM_HOST": "postgres",
        # "TYPEORM_USERNAME": ctx.obj["db"]["admin_user"],
        # "TYPEORM_PASSWORD": ctx.obj["db"]["admin_password"],
        # "TYPEORM_DATABASE": "auth",
        # "TYPEORM_PORT": 5432,
        # "TYPEORM_SYNCHRONIZE": "false",
        # "TYPEORM_LOGGING": "true",
    }
    container = client.containers.run(
        auth_image,
        command,
        remove=False,
        detach=True,
        environment=environment,
        volumes=auth_volumes,
    )
    # exit_code = container.wait()
    logs = client.containers.get(container.id).logs(stream=True, follow=True)

    try:
        while True:
            line = next(logs).decode("utf-8")
            if "Startup completed." in line:
                # todo improve this
                time.sleep(1)
                break
    except StopIteration:
        pass
    except KeyboardInterrupt:
        click.echo("Aborted!")
    # print(logs.decode())
    retry_delays = [1, 5, 10]
    seed_path = os.path.join(
        ctx.obj["install_dir"],
        str(PHTDirectories.SERVICE_DATA_DIR.value),
        "auth",
        "seed.json",
    )
    for i, delay in enumerate(retry_delays):
        try:
            with open(seed_path, "r") as f:
                seed = json.load(f)
            break
        except FileNotFoundError:
            if i == len(retry_delays) - 1:
                raise FileNotFoundError("Seed file not found")
            else:
                click.echo(
                    f"Seed file not found at {seed_path}. Retrying in {delay} seconds.."
                )
                time.sleep(delay)

    robot_id = seed["robotId"]
    robot_secret = seed["robotSecret"]
    # print("".join(output))

    if not (robot_id and robot_secret):
        click.echo(Icons.CROSS.value)
        click.echo("Failed to setup auth server", err=True)
        click.echo(logs, err=True)
        raise Exception("Could not get robot credentials from auth server")

    else:
        auth = {
            "robot_id": robot_id,
            "robot_secret": robot_secret,
        }
        ctx.obj["auth"] = auth

        client.containers.get(container.id).stop()
        click.echo(Icons.CHECKMARK.value)


def write_init_sql(ctx) -> str:
    click.echo("Setting up database... ", nl=False)
    try:
        db_config = ctx.obj["station_config"]["db"]
        init_sql_path = os.path.join(
            ctx.obj["install_dir"],
            str(PHTDirectories.SETUP_SCRIPT_DIR.value),
            "init.sql",
        )
        with open(init_sql_path, "w") as f:
            f.write(templates.render_init_sql(db_user=db_config["admin_user"]))

        click.echo(Icons.CHECKMARK.value)
        # if host path is given, return the path to the init sql file on the host
        if ctx.obj.get("host_path"):
            return os.path.join(
                ctx.obj["host_path"],
                str(PHTDirectories.SETUP_SCRIPT_DIR.value),
                "init.sql",
            )
        return str(init_sql_path)

    except Exception as e:
        click.echo(Icons.CROSS.value)
        click.echo(f"Error creating init sql: {e}", err=True)
        sys.exit(1)


def write_authup_config(ctx):
    authup_path = os.path.join(
        ctx.obj["install_dir"], str(PHTDirectories.CONFIG_DIR.value), "authup"
    )
    os.makedirs(authup_path, exist_ok=True)
    authup_config_path = os.path.join(authup_path, "authup.api.conf")

    auth_config = {
        "admin_password": ctx.obj["station_config"]["admin_password"],
        "port": ctx.obj["station_config"]["auth"]["port"],
        "public_url": "https://"
        + ctx.obj["station_config"]["https"]["domain"]
        + "/auth",
    }

    config = templates.render_authup_api_config(
        auth_config=auth_config,
        db_user=ctx.obj["station_config"]["db"]["admin_user"],
        db_password=ctx.obj["station_config"]["db"]["admin_password"],
    )

    with open(authup_config_path, "w") as f:
        f.write(config)

    authup_config_mount_path = os.path.join(
        _get_base_path(ctx),
        str(PHTDirectories.CONFIG_DIR.value),
        "authup",
        "authup.api.conf",
    )

    return authup_config_mount_path


def write_traefik_configs(ctx) -> Tuple[str, str]:
    click.echo("Setting up traefik... ", nl=False)
    try:
        traefik_config, router_config = templates.render_traefik_configs(
            http_port=ctx.obj["station_config"]["http"]["port"],
            https_port=ctx.obj["station_config"]["https"]["port"],
            https_enabled=True,
            domain=ctx.obj["station_config"]["https"]["domain"],
            certs=ctx.obj["station_config"]["https"]["certificate"],
        )

        traefik_path = os.path.join(
            ctx.obj["install_dir"], str(PHTDirectories.CONFIG_DIR.value), "traefik"
        )

        traefik_config_path = os.path.join(traefik_path, "traefik.yml")
        router_config_path = os.path.join(traefik_path, "config.yml")

        os.makedirs(traefik_path, exist_ok=True)

        with open(traefik_config_path, "w") as f:
            f.write(traefik_config)

        with open(router_config_path, "w") as f:
            f.write(router_config)

        click.echo(Icons.CHECKMARK.value)
        host_path = ctx.obj.get("host_path")
        if host_path:
            traefik_config_path = os.path.join(
                host_path,
                str(PHTDirectories.CONFIG_DIR.value),
                "traefik",
                "traefik.yml",
            )
            router_config_path = os.path.join(
                host_path, str(PHTDirectories.CONFIG_DIR.value), "traefik", "config.yml"
            )
        return str(traefik_config_path), str(router_config_path)

    except Exception as e:
        click.echo(Icons.CROSS.value)
        click.echo(f"Error: {e}", err=True)
        raise e


def write_airflow_config(ctx) -> str:
    click.echo("Setting up airflow... ", nl=False)
    try:
        db_connection_string = (
            f"postgresql+psycopg2://{ctx.obj['station_config']['db']['admin_user']}:{ctx.obj['station_config']['db']['admin_password']}"
            f"@postgres/airflow"
        )
        airflow_config = templates.render_airflow_config(
            sql_alchemy_conn=db_connection_string,
            domain=ctx.obj["station_config"]["https"]["domain"],
        )

        path = _get_base_path(ctx)

        airflow_config_path = os.path.join(
            path, str(PHTDirectories.CONFIG_DIR.value), "airflow.cfg"
        )

        write_path = os.path.join(
            ctx.obj["install_dir"], str(PHTDirectories.CONFIG_DIR.value), "airflow.cfg"
        )

        with open(write_path, "w") as f:
            f.write(airflow_config)

        click.echo(Icons.CHECKMARK.value)
        return str(airflow_config_path)

    except Exception as e:
        click.echo(Icons.CROSS.value)
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)


def write_compose_file(ctx):
    host_path = ctx.obj.get("host_path")
    if host_path:
        host_path = os.path.join(ctx.obj["host_path"], "docker-compose.yml")

    compose_path = os.path.join(ctx.obj["install_dir"], "docker-compose.yml")
    click.echo(
        f"Writing compose file to {host_path if host_path else compose_path}... ",
        nl=False,
    )

    content = templates.render_compose(ctx=ctx.obj)
    with open(compose_path, "w") as f:
        f.write(content)

    click.echo(Icons.CHECKMARK.value)


def _get_base_path(ctx) -> str:
    host_path = ctx.obj.get("host_path")
    if host_path:
        path = host_path
    else:
        path = ctx.obj["install_dir"]

    return path



import os

from jinja2 import Environment, FileSystemLoader, PackageLoader


def get_template_env():
    env_template_dir = os.getenv("PHT_TEMPLATE_DIR")
    if env_template_dir:
        loader = FileSystemLoader(env_template_dir)
    else:
        loader = PackageLoader("station.ctl", "templates")
    return Environment(loader=loader)



from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from station.app.schemas import datasets


class DBSchema(BaseModel):
    class Config:
        orm_mode = True


class TrainState(DBSchema):
    id: int
    iteration: int
    round: int
    updated_at: Optional[datetime] = None
    signing_key: Optional[str] = None
    sharing_key: Optional[str] = None
    seed: Optional[int] = None
    key_broadcast: Optional[str] = None


class TrainCreate(BaseModel):
    name: Optional[str] = None
    proposal_id: Optional[str] = None


class TrainUpdate(TrainCreate):
    pass


class Train(DBSchema):
    id: int
    name: Optional[str] = None
    created_at: datetime
    updated_at: Optional[datetime] = None
    is_active: bool
    proposal_id: Optional[str] = None
    state: TrainState
    token: Optional[str] = None
    dataset: Optional[datasets.DataSet] = None


class FederatedTrainConfigCreate(DBSchema):
    name: str



import functools
import json
import os
import re
from enum import Enum
from typing import Optional, Tuple, Union

from cryptography.fernet import Fernet
from dotenv import find_dotenv, load_dotenv
from loguru import logger
from pydantic import AnyHttpUrl, AnyUrl, BaseModel, PostgresDsn, SecretStr, parse_obj_as
from yaml import safe_dump, safe_load

from station.app.env import StationEnvironmentVariables


class Emojis(str, Enum):
    """
    Enum of emojis to use in the app.
    """

    INFO = "ℹ️"
    SUCCESS = "✅"
    WARNING = "⚠️"
    ERROR = "❌"


class RegistrySettings(BaseModel):
    address: Union[AnyHttpUrl, str]
    user: str
    password: SecretStr
    project: Optional[str] = None

    @property
    def api_url(self) -> str:
        """
        Returns the API URL of the registry.
        """
        if isinstance(self.address, str):
            if self.address.startswith("https://"):
                return self.address + "/api/v2.0"
            else:
                return "https://" + self.address + "/api/v2.0"
        else:
            return self.address + "/api/v2.0"


class AirflowSettings(BaseModel):
    host: Optional[Union[AnyHttpUrl, str]] = "airflow"
    # api_url: Optional[str] = "http://localhost:8080/api/v1/"
    port: Optional[int] = None
    user: Optional[str] = "admin"
    password: Union[SecretStr, str]
    station_db_dsn: Optional[PostgresDsn] = None

    @property
    def api_url(self) -> str:
        """
        Returns the API URL of Airflow.
        """
        if isinstance(self.host, str):
            if self.host.startswith("http://") or self.host.startswith("https://"):
                api_url = self.host + (f":{self.port}" if self.port else "")
            else:
                api_url = f"http://{self.host}" + f":{self.port}" if self.port else ""

        elif isinstance(self.host, AnyHttpUrl):
            url = "http://" + self.host.host + f":{self.port}" if self.port else ""
            api_url = url + self.host.path if self.host.path else ""
        else:
            raise ValueError("Airflow host is not a valid URL.")
        if not api_url.endswith("/api/v1/"):
            if api_url.endswith("/"):
                api_url += "api/v1/"
            else:
                api_url += "/api/v1/"
        return api_url


class MinioSettings(BaseModel):
    host: Optional[Union[AnyHttpUrl, AnyUrl, str]] = "minio"
    port: Optional[int] = None
    access_key: str
    secret_key: Union[SecretStr, str]

    @property
    def server_url(self) -> str:
        """
        Returns the server URL of Minio.
        """
        return f"{self.host}{':' + str(self.port) if self.port else ''}"


class CentralUISettings(BaseModel):
    api_url: Optional[AnyHttpUrl] = None
    robot_id: Optional[str] = None
    robot_secret: Optional[SecretStr] = None


class RedisSettings(BaseModel):
    host: Optional[str] = "redis"
    port: Optional[int] = 6379
    password: Optional[SecretStr] = None
    db: Optional[int] = 0


class AuthConfig(BaseModel):
    robot_id: str
    robot_secret: SecretStr
    host: Optional[Union[AnyHttpUrl, AnyUrl, str]] = "station-auth"
    port: Optional[int] = None

    @property
    def token_url(self) -> str:
        if self.host.startswith("http") or self.host.startswith("https"):
            return f"{self.host}{f':{self.port}' if self.port else ''}/token"
        return f"http://{self.host}{f':{self.port}' if self.port else ''}/token"

    @property
    def user_url(self) -> str:
        if self.host.startswith("http") or self.host.startswith("https"):
            return f"{self.host}{f':{self.port}' if self.port else ''}/users"
        return f"http://{self.host}{f':{self.port}' if self.port else ''}/users"

    @property
    def auth_url(self) -> Union[AnyUrl, str]:
        if self.host.startswith("http") or self.host.startswith("https"):
            return f"{self.host}{f':{self.port}' if self.port else ''}"
        return f"http://{self.host}{f':{self.port}' if self.port else ''}"


class StationRuntimeEnvironment(str, Enum):
    """
    Enum of the runtime environments of the station.
    """

    DEVELOPMENT = "development"
    PRODUCTION = "production"
    TESTING = "testing"


class DatabaseSettings(BaseModel):
    admin_user: str
    admin_password: str
    host: Optional[str] = "postgres"
    port: Optional[int] = 5432
    database: Optional[str] = "pht_station"

    @property
    def dsn(self) -> PostgresDsn:
        return PostgresDsn(
            scheme="postgresql+psycopg2",
            user=self.admin_user,
            password=self.admin_password,
            url=self.host,
            port=self.port,
            path=f"/{self.database}",
        )

    @classmethod
    def from_dsn(cls, dsn: PostgresDsn):
        return cls(
            admin_user=dsn.user,
            admin_password=dsn.password,
            host=dsn.host,
            port=dsn.port,
            database=dsn.path,
        )


class StationConfig(BaseModel):
    """
    Object containing the configuration of the station.
    """

    station_id: Union[int, str]
    station_data_dir: Optional[str]
    host: Optional[Union[AnyHttpUrl, str]] = os.getenv(
        StationEnvironmentVariables.STATION_API_HOST.value, "0.0.0.0"
    )
    port: Optional[int] = os.getenv(
        StationEnvironmentVariables.STATION_API_PORT.value, 8000
    )
    db: DatabaseSettings
    environment: Optional[
        StationRuntimeEnvironment
    ] = StationRuntimeEnvironment.DEVELOPMENT
    fernet_key: Optional[SecretStr] = None
    registry: RegistrySettings
    auth: Optional[AuthConfig] = None
    airflow: Optional[AirflowSettings] = None
    minio: Optional[MinioSettings] = None
    central_ui: Optional[CentralUISettings] = CentralUISettings()
    redis: Optional[RedisSettings] = RedisSettings()

    @classmethod
    def from_file(cls, path: str) -> "StationConfig":
        """
        Parse the application configuration file (station_config.yml) for the settings required to run the station api.
        Args:
            path: location of the configuration file

        Returns:
            StationConfig object containing the configuration of the station api.
        """

        return cls._parse_config_file(path)

    @staticmethod
    def _parse_config_file(path: str) -> "StationConfig":
        with open(path, "r") as f:
            config_dict = safe_load(f)

        "postgresql+psycopg2://{}:{}@{}:{}/{}".format(
            config_dict["db"]["admin_user"],
            config_dict["db"]["admin_password"],
            config_dict["db"].get("host", "postgres"),
            config_dict["db"].get("port", 5432),
            config_dict["db"].get("database", "pht_station"),
        )

        db = DatabaseSettings(
            admin_user=config_dict["db"]["admin_user"],
            admin_password=config_dict["db"]["admin_password"],
            host=config_dict["db"].get("host", "postgres"),
            port=config_dict["db"].get("port", 5432),
            database=config_dict["db"].get("database", "pht_station"),
        )

        registry_settings = RegistrySettings(
            address=config_dict["registry"]["address"],
            user=config_dict["registry"]["user"],
            password=config_dict["registry"]["password"],
            project=config_dict["registry"].get("project", None),
        )

        auth_settings = AuthConfig(
            robot_id=config_dict["auth"]["robot_id"],
            robot_secret=config_dict["auth"]["robot_secret"],
            host=config_dict["auth"].get("host", "station-auth"),
            port=config_dict["auth"].get("port", 3010),
        )

        airflow_settings = AirflowSettings(
            host=config_dict.get("airflow").get("host", "airflow"),
            port=config_dict.get("airflow").get("port", 8080),
            user=config_dict.get("airflow").get("admin_user", "admin"),
            password=config_dict.get("airflow").get("admin_password"),
        )
        # airflow_settings.api_url = f"http://{airflow_settings.host}:{airflow_settings.port}/api/v1/"

        minio_settings = MinioSettings(
            host=config_dict.get("minio").get("host", "minio"),
            port=config_dict.get("minio").get("port", 9000),
            access_key=config_dict.get("minio").get("admin_user", "minio_admin"),
            secret_key=config_dict.get("minio").get("admin_password"),
        )

        central_settings = CentralUISettings(
            api_url=config_dict.get("central").get("api_url"),
            robot_id=config_dict.get("central").get("robot_id"),
            robot_secret=config_dict.get("central").get("robot_secret"),
        )

        return StationConfig(
            station_id=config_dict["station_id"],
            station_data_dir=config_dict["station_data_dir"],
            host=os.getenv(
                str(StationEnvironmentVariables.STATION_API_HOST.value), "0.0.0.0"
            ),
            port=os.getenv(
                str(StationEnvironmentVariables.STATION_API_PORT.value), 8000
            ),
            db=db,
            environment=StationRuntimeEnvironment(config_dict["environment"]),
            fernet_key=config_dict["api"]["fernet_key"],
            registry=registry_settings,
            auth=auth_settings,
            airflow=airflow_settings,
            minio=minio_settings,
            central_ui=central_settings,
            redis=RedisSettings(),
        )

    def to_file(self, path: str) -> None:
        # todo get secret values recursively
        with open(path, "w") as f:
            safe_dump(json.loads(self.json(indent=2)), f)


# https://stackoverflow.com/a/31174427/3838313
def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))


# Evaluation for initialization of values file < environment
class Settings:
    """
    Class to handle the settings of the station API and connections to other services. Can be configured via a file or
    environment variables.
    """

    config: StationConfig
    config_path: Optional[str]
    is_initialized: bool

    def __init__(self, config_path: str = None, config: StationConfig = None):
        load_dotenv(find_dotenv())
        self._config_file = False
        self.config = None
        self.is_initialized = False
        if config:
            self.config = config
        elif config_path:
            self.config_path = config_path
            self._config_file = True
        else:
            self.config_path = os.getenv(
                StationEnvironmentVariables.CONFIG_PATH.value, "station_config.yml"
            )
            self._config_file = True
        # todo create/update config file
        if os.getenv("ENVIRONMENT") == "development":
            self.setup()

    def setup(self) -> StationConfig:
        """
        Initialize the settings. First tries to load the settings from the config file. Then tries to load the settings
        from the environment variables. If a conflicting value is found the environment variable has precedence. If both
        fail, the default settings are used.

        Returns:
            StationConfig The initialized settings.

        """
        logger.info(f"{Emojis.INFO.value}Setting up station backend...")
        # check for config file at given or default path and load if exists
        self._check_config_path()
        # validate the runtime environment
        self._setup_runtime_environment()
        self._setup_station_environment()

        logger.info(f"Station backend setup successful {Emojis.SUCCESS.value}")
        self.is_initialized = True
        return self.config

    def get_fernet(self) -> Fernet:
        """
        Get the Fernet key for encryption and decryption of secrets. Configured via environment variables or station
        config.

        Returns: A fernet object that can be used to encrypt and decrypt secrets.
        """
        if self.config.fernet_key is None:
            raise ValueError("No Fernet key provided")

        key = str(self.config.fernet_key)
        return Fernet(key.encode())

    def _check_config_path(self):
        """
        Checks if a config is present at the given path. Raise an error if custom config file is given but not present.

        Returns:

        """
        print(os.getcwd())
        logger.info(f"{Emojis.INFO.value}Looking for config file...")
        print("Config path", self.config_path)
        if not os.path.isfile(self.config_path):
            if self.config_path == "station_config.yml":
                logger.warning(
                    f"\t{Emojis.WARNING.value} "
                    f"No config file found. Attempting configuration via environment variables..."
                )
            else:
                raise FileNotFoundError(
                    f"{Emojis.ERROR.value}   Custom config file not found at {self.config_path}."
                )
            # construct a placeholder config to fill with environment and default values
            self.config = StationConfig.construct()
        # Parse the config file
        else:
            logger.info(
                f"\t{Emojis.SUCCESS.value}   Config file found at {self.config_path} loading... "
            )
            config = StationConfig.from_file(self.config_path)
            print(config)
            self.config = config
            logger.info(f"{Emojis.SUCCESS.value} Config loaded.")

    def _setup_runtime_environment(self):
        """
        Setup and validate the runtime environment (development|production) of the station API
        Returns:

        """
        environment_var = os.getenv(StationEnvironmentVariables.ENVIRONMENT.value)
        # if config variable is in environment variables use it
        if environment_var:
            try:
                runtime_environment = StationRuntimeEnvironment(environment_var)
                # display when config variable is overwritten with env var
                if self.config.environment:
                    logger.debug(
                        f"{Emojis.INFO.value}Overriding runtime environment with env var specification."
                    )
            except ValueError:
                raise ValueError(
                    f"{Emojis.ERROR.value}   Invalid value ({environment_var}) for runtime environment"
                    f" in env var {StationEnvironmentVariables.ENVIRONMENT.value}."
                )
        # otherwise, use the value parse from config file
        else:
            runtime_environment = self.config.environment

        # Display runtime environment
        if runtime_environment == StationRuntimeEnvironment.PRODUCTION:
            logger.info(f"{Emojis.INFO.value}Running in production environment.")
        elif runtime_environment == StationRuntimeEnvironment.DEVELOPMENT:
            logger.warning(
                f"{Emojis.WARNING.value}Development environment detected,"
                f" set environment variable '{StationEnvironmentVariables.ENVIRONMENT.value}' "
                f"to '{StationRuntimeEnvironment.PRODUCTION.value}' for production mode."
            )
        # set the parsed runtime environment
        self.config.environment = runtime_environment

    def _setup_station_environment(self):
        """
        After potential config files are found, parse the environment variables to override the config file values if
        they exist.
        Validate the configuration from both options.

        Returns:

        """

        # ensure that the station id is set
        env_station_id = os.getenv(StationEnvironmentVariables.STATION_ID.value)

        if not env_station_id:
            try:
                # todo improve this
                pass
            except AttributeError:
                raise ValueError(
                    f"{Emojis.ERROR.value}   No station id specified in config or env var "
                    f"{StationEnvironmentVariables.STATION_ID.value}."
                )
        elif env_station_id:
            logger.debug(
                f"\t{Emojis.INFO.value}Overriding station id with env var specification."
            )
            self.config.station_id = env_station_id

        # parse environment variables
        self._setup_station_api()
        self._setup_central_api_connection()
        self._setup_airflow()
        self._setup_fernet()
        self._setup_station_auth()
        self._setup_redis()
        self._setup_registry_connection()
        self._setup_minio_connection()

    def _setup_station_api(self):
        """
        Configure the station api url, port and database connection from environment variables or config file.
        Returns:

        """
        # Try to find api host and port in environment variables
        env_station_host = os.getenv(StationEnvironmentVariables.STATION_API_HOST.value)
        if env_station_host:
            logger.debug(
                f"\t{Emojis.INFO.value}Overriding station api host with env var specification."
            )
            self.config.host = env_station_host
        env_station_port = os.getenv(StationEnvironmentVariables.STATION_API_PORT.value)
        if env_station_port:
            logger.debug(
                f"\t{Emojis.INFO.value}Overriding station api port with env var specification."
            )
            self.config.port = int(env_station_port)
        station_db = os.getenv(StationEnvironmentVariables.STATION_DB.value)
        if station_db:
            logger.debug(
                f"\t{Emojis.INFO.value}Overriding station db with env var specification."
            )
            self.config.db = DatabaseSettings.from_dsn(
                parse_obj_as(PostgresDsn, station_db)
            )
        else:
            if self.config.environment == "production":
                raise ValueError(
                    f"{Emojis.ERROR.value} Connection string to station database needs to be specified in"
                    f" environment variables."
                )
            else:
                logger.warning(
                    f"{Emojis.WARNING.value} Connection string to station database is not specified in"
                    f" environment variables. Default database is used."
                )
        station_data_dir = os.getenv(StationEnvironmentVariables.STATION_DATA_DIR.value)
        if station_data_dir:
            self.config.station_data_dir = station_data_dir

        # if "sqlite" in self.config.db.dsn.lower():
        #     if self.config.environment == StationRuntimeEnvironment.PRODUCTION:
        #         raise ValueError(f"{Emojis.ERROR.value}   SQLite database not supported for production mode.")
        #     else:
        #         logger.warning(f"{Emojis.WARNING.value}   SQLite database only supported in development mode.")

    def _setup_central_api_connection(self):
        self._validate_config_item(
            env_var=StationEnvironmentVariables.CENTRAL_API_URL,
            config_item="central_ui.api_url",
        )
        self._validate_config_item(
            env_var=StationEnvironmentVariables.STATION_ROBOT_ID,
            config_item="central_ui.robot_id",
        )
        self._validate_config_item(
            env_var=StationEnvironmentVariables.STATION_ROBOT_SECRET,
            config_item="central_ui.robot_secret",
        )

        logger.info(
            f"{Emojis.INFO.value} Central API url: {self.config.central_ui.api_url}, "
            f"Robot ID: {self.config.central_ui.robot_id} Robot Secret: ******"
        )

    def _validate_config_item(
        self,
        env_var: StationEnvironmentVariables,
        config_item: str,
        error_message: str = None,
        warning_message: str = None,
        item_name: str = None,
    ):
        """
        Validates that a config item is present in the environment variables. If not, a warning or error is raised.
        Args:
            env_var: the environment variable the item can be found in
            config_item: attr path of the station configuration of the item
            error_message: optional custom error message
            warning_message: optional custom warning message
            item_name: optional display name of the item in the error messages

        Returns:

        """
        env_var = os.getenv(env_var.value)
        if env_var:
            logger.debug(
                f"\t{Emojis.INFO.value}Overriding {item_name if item_name else config_item} with env var specification."
            )
            rsetattr(self.config, config_item, env_var)
        elif rgetattr(self.config, config_item):
            logger.debug(
                f"\t{Emojis.INFO.value}{item_name if item_name else config_item}: {rgetattr(self.config, config_item)}"
            )
        else:
            if self.config.environment == "production":
                if error_message:
                    raise ValueError(f"{Emojis.ERROR.value}{error_message}")
                else:
                    raise ValueError(
                        f"{Emojis.ERROR.value} {item_name if item_name else config_item} needs to be specified."
                    )
            else:
                if warning_message:
                    logger.warning(f"{Emojis.WARNING.value}{warning_message}")
                else:
                    logger.warning(
                        f"{Emojis.WARNING.value} {item_name if item_name else config_item} is not specified."
                    )

    def _setup_fernet(self):
        """
        Configure the fernet key from environment variables or config file.
        In the development mode if none is given generate a new one.
        Raise an error if no key is given in production mode.
        Returns:

        """
        env_fernet_key = os.getenv(str(StationEnvironmentVariables.FERNET_KEY.value))

        if not env_fernet_key and not self.config.fernet_key:
            # if fernet key is given in production raise an error
            if self.config.environment == StationRuntimeEnvironment.PRODUCTION:
                raise ValueError(
                    f"{Emojis.ERROR.value}   No fernet key specified in config or env vars."
                )
            # generate new key in development mode
            elif self.config.environment == StationRuntimeEnvironment.DEVELOPMENT:
                logger.warning(
                    f"\t{Emojis.WARNING.value}No fernet key specified in config or env vars"
                )
                self.config.fernet_key = Fernet.generate_key().decode()
                logger.info(
                    f"\t{Emojis.INFO.value}Generated new key for development environment."
                )
        elif env_fernet_key and self.config.fernet_key:
            logger.debug(
                f"\t{Emojis.INFO.value}Overriding fernet key with env var specification."
            )
            self.config.fernet_key = env_fernet_key

        elif env_fernet_key:
            self.config.fernet_key = env_fernet_key

    def _setup_redis(self):
        """
        Configure the redis connection from environment variables or config file.
        Returns:

        """

        logger.info(f"{Emojis.INFO.value}Setting up redis connection...")
        redis_config = self.config.redis.copy()

        host = os.getenv(str(StationEnvironmentVariables.REDIS_HOST.value))
        port = os.getenv(str(StationEnvironmentVariables.REDIS_PORT.value))
        db = os.getenv(str(StationEnvironmentVariables.REDIS_DB.value))
        password = os.getenv(str(StationEnvironmentVariables.REDIS_PW.value))

        if host:
            logger.debug(
                f"\t{Emojis.INFO.value}Overriding redis connections with env var specification."
            )
            if not redis_config:
                redis_config = RedisSettings()
            redis_config.host = host
            redis_config.port = port or redis_config.port
            redis_config.db = db
            redis_config.password = password
            self.config.redis = redis_config
        if not self.config.redis:
            self.config.redis = RedisSettings()
            logger.info(
                f"\t{Emojis.INFO.value}No redis connection specified in config or env vars. Using default."
                f" Host: {self.config.redis.host}, Port: {self.config.redis.port}"
            )

    def _setup_station_auth(self):
        """
        Configure the connection to the station auth from environment variables or config file.
        Optional in development mode, raises an error if not configured in production mode.
        Returns:

        """
        logger.info(f"{Emojis.INFO.value}Setting up station authentication...")
        # check if station auth is configured via environment variables
        (
            env_auth_server,
            env_auth_port,
            env_auth_robot,
            env_auth_robot_secret,
        ) = self._get_internal_service_env_vars(
            host=StationEnvironmentVariables.AUTH_SERVER_HOST,
            port=StationEnvironmentVariables.AUTH_SERVER_PORT,
            user=StationEnvironmentVariables.AUTH_ROBOT_ID,
            secret=StationEnvironmentVariables.AUTH_ROBOT_SECRET,
        )

        _auth_server = False
        # ensure there is an auth server specified in production mode
        if not (
            (env_auth_server and env_auth_robot and env_auth_robot_secret)
            or self.config.auth
        ):
            if self.config.environment == StationRuntimeEnvironment.PRODUCTION:
                raise ValueError(
                    f"{Emojis.ERROR.value}   No station auth specified in config or env vars,"
                    f" invalid configuration for production"
                )

        # no auth config present at initialization, create a new one
        elif not self.config.auth and (
            env_auth_server and env_auth_robot and env_auth_robot_secret
        ):
            logger.debug(
                f"{Emojis.INFO.value}No auth config found, creating new one from env vars."
            )
            env_auth_config = AuthConfig(
                host=env_auth_server,
                port=env_auth_port,
                robot_id=env_auth_robot,
                robot_secret=SecretStr(env_auth_robot_secret),
            )

            self.config.auth = env_auth_config
            _auth_server = True
        # if config and env vars exist override the config values with the environment variables
        elif self.config.auth and (
            env_auth_server or env_auth_robot or env_auth_robot_secret or env_auth_port
        ):
            logger.debug("Overriding auth server config with env var specifications.")
            if env_auth_server:
                self.config.auth.host = env_auth_server
            if env_auth_port:
                self.config.auth.port = env_auth_port
            if env_auth_robot:
                self.config.auth.robot_id = env_auth_robot
            if env_auth_robot_secret:
                self.config.auth.robot_secret = env_auth_robot_secret
            # validate the overridden config
            self.config.auth = AuthConfig(**self.config.auth.dict())
            _auth_server = True

        # log authentication server config and status
        if _auth_server:
            logger.info(
                f"Auth server: url - {self.config.auth.host}:{self.config.auth.port},"
                f" robot - {self.config.auth.robot_id}"
            )
        else:
            logger.warning(
                f"{Emojis.WARNING.value}No auth server specified in config or env vars,"
                f" ignoring in development mode"
            )

    def _setup_registry_connection(self):
        """
        Configure the connection to the central container registry from environment variables or config file.
        Raises an error if not configured in production mode.
        Returns:

        """
        logger.info(f"{Emojis.INFO.value}Setting up registry connection...")
        (
            env_registry,
            env_registry_user,
            env_registry_password,
        ) = self._get_external_service_env_vars(
            url=StationEnvironmentVariables.REGISTRY_URL,
            client_id=StationEnvironmentVariables.REGISTRY_USER,
            client_secret=StationEnvironmentVariables.REGISTRY_PW,
        )

        # catch attribute error if no registry config is present in config
        try:
            registry_config = self.config.registry
        except AttributeError:
            registry_config = None
        _registry_config = False
        # override registry config if env vars are present and validate afterwards
        if registry_config and (
            env_registry and env_registry_user and env_registry_password
        ):
            logger.debug("Overriding registry config with env var specifications.")
            if env_registry:
                registry_config.address = env_registry
            if env_registry_user:
                registry_config.user = env_registry_user
            if env_registry_password:
                registry_config.password = env_registry_password
            # validate the overridden config
            registry_config = RegistrySettings(**registry_config.dict())
        # no registry config found
        elif not registry_config and not (
            env_registry and env_registry_user and env_registry_password
        ):
            _registry_config = False
        elif not registry_config and (
            env_registry and env_registry_user and env_registry_password
        ):
            logger.debug(
                f"{Emojis.INFO.value}No registry config found, creating new one from env vars."
            )
            registry_config = RegistrySettings(
                address=env_registry,
                user=env_registry_user,
                password=env_registry_password,
            )
            self.config.registry = registry_config
            self._validate_config_item(
                StationEnvironmentVariables.REGISTRY_PROJECT,
                "registry.project",
            )
        # log registry config and status
        if registry_config:
            logger.info(
                f"Registry: url - {self.config.registry.address}, user - {self.config.registry.user}, "
                f"project - {self.config.registry.project}"
            )
        else:
            # raise error if no registry is configured in production mode
            if self.config.environment == StationRuntimeEnvironment.PRODUCTION:
                raise ValueError(
                    f"{Emojis.ERROR.value}   No registry config specified in config or env vars"
                )
            else:
                logger.warning(
                    "No registry config specified in config or env vars, ignoring in development mode"
                )

    def _setup_minio_connection(self):
        """
        Configure the connection to the minio storage from environment variables or config file.
        Raises an error if not configured in production mode.
        Returns:

        """
        logger.info("Setting up minio connection...")
        # get the environment variables for minio
        env_connection = self._get_internal_service_env_vars(
            host=StationEnvironmentVariables.MINIO_HOST,
            port=StationEnvironmentVariables.MINIO_PORT,
            user=StationEnvironmentVariables.MINIO_ACCESS_KEY,
            secret=StationEnvironmentVariables.MINIO_SECRET_KEY,
        )
        (
            env_minio_host,
            env_minio_port,
            env_minio_access_key,
            env_minio_secret_key,
        ) = env_connection

        # get minio from config file or construct dummy
        _minio_config = False
        minio_config = self.config.minio
        if minio_config:
            _minio_config = True
            logger.debug("Minio config found, checking for env vars.")
        else:
            minio_config = MinioSettings.construct()

        # override minio config if config and env vars are present and validate afterwards
        if _minio_config and (
            env_minio_host
            or env_minio_port
            or env_minio_access_key
            or env_minio_secret_key
        ):
            logger.debug("Overriding minio config with env var specifications.")
            if env_minio_host:
                minio_config.host = env_minio_host
            if env_minio_port:
                minio_config.port = env_minio_port
            if env_minio_access_key:
                minio_config.access_key = env_minio_access_key
            if env_minio_secret_key:
                minio_config.secret_key = env_minio_secret_key
            _minio_config = True
        # no minio config found
        elif not _minio_config and not (
            env_minio_host and env_minio_access_key and env_minio_secret_key
        ):
            _minio_config = False

        # no config but environment variables are found
        elif not _minio_config and (
            env_minio_host and env_minio_access_key and env_minio_secret_key
        ):
            logger.debug(
                f"{Emojis.INFO.value}No minio config found, creating new one from env vars."
            )
            minio_config.host = env_minio_host
            minio_config.access_key = env_minio_access_key
            minio_config.secret_key = env_minio_secret_key
            if env_minio_port:
                minio_config.port = env_minio_port
            _minio_config = True

        # log minio config and status
        if _minio_config:
            # validate the overridden config
            self.config.minio = MinioSettings(**minio_config.dict())
            logger.info(
                f"Minio: host - {self.config.minio.host}, port - {self.config.minio.port}"
            )
        else:
            # raise error if no minio is configured in production mode
            if self.config.environment == StationRuntimeEnvironment.PRODUCTION:
                raise ValueError(
                    f"{Emojis.ERROR}   No minio config specified in config or env vars"
                )
            else:
                logger.warning(
                    "No minio config specified in config or env vars, ignoring in development mode"
                )

    def _setup_airflow(self):
        logger.info(
            "Setting up airflow connection and airflow-connection to station database..."
        )

        # get the environment variables for airflow
        (
            env_airflow_api_url,
            env_airflow_port,
            env_airflow_user,
            env_airflow_secret,
        ) = self._get_internal_service_env_vars(
            host=StationEnvironmentVariables.AIRFLOW_API_URL,
            port=StationEnvironmentVariables.AIRFLOW_PORT,
            user=StationEnvironmentVariables.AIRFLOW_USER,
            secret=StationEnvironmentVariables.AIRFLOW_PW,
        )

        # get airflow from config file or construct dummy
        _airflow_config = False
        airflow_config = self.config.airflow
        if airflow_config:
            _airflow_config = True
            logger.debug("Airflow config found, checking for env vars.")
        else:
            airflow_config = AirflowSettings.construct()

        # override airflow config if config and env vars are present and validate afterwards
        if _airflow_config and (
            env_airflow_api_url
            or env_airflow_port
            or env_airflow_user
            or env_airflow_secret
        ):
            logger.debug("Overriding airflow config with env var specifications.")
            if env_airflow_api_url:
                airflow_config.host = env_airflow_api_url
            if env_airflow_port:
                airflow_config.port = env_airflow_port
            if env_airflow_user:
                airflow_config.user = env_airflow_user
            if env_airflow_secret:
                airflow_config.password = env_airflow_secret
            _airflow_config = True
        # no airflow config found
        elif not _airflow_config and not (
            env_airflow_api_url and env_airflow_user and env_airflow_secret
        ):
            _airflow_config = False

        # no config but environment variables are found
        elif not _airflow_config and (
            env_airflow_api_url and env_airflow_user and env_airflow_secret
        ):
            logger.debug(
                f"{Emojis.INFO.value}No airflow config found, creating new one from env vars."
            )
            airflow_config.host = env_airflow_api_url.split("://")[1]
            if re.search(r":\d+", env_airflow_api_url):
                airflow_config.port = (
                    re.search(r":\d+", env_airflow_api_url).group(0).replace(":", "")
                )
            airflow_config.user = env_airflow_user
            airflow_config.password = env_airflow_secret
            if env_airflow_port:
                airflow_config.port = env_airflow_port
            _airflow_config = True

        # log airflow config and status
        if _airflow_config:
            # validate the overridden config
            self.config.airflow = AirflowSettings(**airflow_config.dict())
            logger.info(
                f"Airflow: API url - {self.config.airflow.api_url}, port - {self.config.airflow.port}"
            )
        else:
            # raise error if no airflow is configured in production mode
            if self.config.environment == StationRuntimeEnvironment.PRODUCTION:
                raise ValueError(
                    f"{Emojis.ERROR.value}   No airflow config specified in config or env vars"
                )
            else:
                logger.warning(
                    "No airflow config specified in config or env vars, generating defaults for "
                    "development mode"
                )
                self.config.airflow = AirflowSettings()

        # todo re-enable this when checked if it is necessary -> should probably just use api calls to station api
        # # Check whether connection to station database with connection_id already exists, if not create it
        # try:
        #     self._create_station_db_connection()
        # except Exception as e:
        #     logger.error(f"{Emojis.ERROR}   Could not create airflow connection to station database: {e}")
        #     if self.config.environment == StationRuntimeEnvironment.PRODUCTION:
        #         raise ValueError(f"{Emojis.ERROR.value}   Unable to add database connection to airflow")
        #     else:
        #         logger.warning(f"Unable to add database connection to airflow. Is airflow running?")
        #         logger.error(e)

    # def _create_station_db_connection(self):
    #
    #     connection_id = "pg_station"
    #     self.config.airflow.station_db_dsn = self.config.db.dsn
    #     if isinstance(self.config.db, str):
    #         credentials, host = self.config.db.split("://")[1].split("@")
    #     elif isinstance(self.config.db, SecretStr):
    #         credentials, host = (
    #             self.config.db.get_secret_value().split("://")[1].split("@")
    #         )
    #     else:
    #         raise ValueError(
    #             f"{Emojis.ERROR.value.value}   Unable to parse database connection string"
    #         )
    #     user, password = credentials.split(":")
    #     host, schema = host.split("/")
    #     port = 5432
    #     if len(host.split(":")) == 2:
    #         host, port = host.split(":")
    #         port = int(port)
    #
    #     conn = {
    #         "connection_id": connection_id,
    #         "conn_type": "postgres",
    #         "host": "postgres",
    #         "login": user,
    #         "port": port,
    #         "password": password,
    #         "schema": schema,
    #     }
    #
    #     # Check whether connection with connection_id already exists, if not create it
    #     url_get = self.config.airflow.api_url + f"connections/{connection_id}"
    #     url_post = self.config.airflow.api_url + "connections"
    #     auth = HTTPBasicAuth(
    #         self.config.airflow.user, self.config.airflow.password.get_secret_value()
    #     )
    #     r = requests.get(url=url_get, auth=auth, verify=False)
    #
    #     if r.status_code != 200:
    #         logger.debug(
    #             f"\t{Emojis.INFO}Database connection in airflow with connection id {connection_id} does not exist,"
    #             f" creating new one from environment variables."
    #         )
    #         try:
    #             r = requests.post(url=url_post, auth=auth, json=conn, verify=False)
    #             r.raise_for_status()
    #             logger.info(
    #                 f"\t{Emojis.INFO} Database connection in airflow with id {connection_id} got created."
    #             )
    #         except Exception as e:
    #             f"\t{Emojis.WARNING}Error occured while trying to create the database connection "
    #             f"in airflow with id " f"{connection_id}."
    #             f"\t{Emojis.WARNING} -- {e}."
    #     else:
    #         logger.info(
    #             f"\t{Emojis.INFO} Database connection in airflow with id {connection_id} exists."
    #        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(config={self.config})"

    @staticmethod
    def _get_external_service_env_vars(
        url: StationEnvironmentVariables,
        client_id: StationEnvironmentVariables,
        client_secret: StationEnvironmentVariables,
    ) -> Tuple[str, str, str]:
        """
        Get the tuple of connection variables for an external service from environment variables.
        Args:
            url:
            client_id:
            client_secret:

        Returns:

        """
        env_url = os.getenv(url.value)
        env_client_id = os.getenv(client_id.value)
        env_client_secret = os.getenv(client_secret.value)
        return env_url, env_client_id, env_client_secret

    @staticmethod
    def _get_internal_service_env_vars(
        host: StationEnvironmentVariables,
        port: StationEnvironmentVariables,
        user: StationEnvironmentVariables,
        secret: StationEnvironmentVariables,
    ) -> Tuple[str, int, str, str]:
        """
        Get the tuple of connection variables for an internal service from environment variables.
        Args:
            host:
            port:
            user:
            secret:

        Returns:

        """
        env_server_host = os.getenv(host.value)
        env_server_port = os.getenv(port.value)
        if env_server_port:
            env_server_port = int(env_server_port)
        env_server_user = os.getenv(user.value)
        env_server_secret = os.getenv(secret.value)
        return env_server_host, env_server_port, env_server_user, env_server_secret


settings = Settings()



import os
from typing import Optional, Union

import yaml
from pydantic import AnyHttpUrl, BaseSettings, SecretStr, validator
from pydantic.env_settings import SettingsSourceCallable
from rich.pretty import pprint

from station.common.config.fix import ConfigItemFix
from station.common.config.validators import (
    admin_validator,
    dir_validator,
    file_readable_validator,
    validate_fernet_key,
    validate_file_readable,
)
from station.common.constants import ApplicationEnvironment


class StationSettingsConfig:
    """Shared configuration for station settings"""

    use_enum_values = True
    env_prefix: str

    @classmethod
    def with_prefix(cls, prefix: str):
        """Returns a new class with the given prefix"""
        return type(cls.__name__, (cls,), {"env_prefix": prefix})

    @classmethod
    def customise_sources(
        cls,
        init_settings: SettingsSourceCallable,
        env_settings: SettingsSourceCallable,
        file_secret_settings: SettingsSourceCallable,
    ) -> tuple[SettingsSourceCallable, ...]:
        return env_settings, init_settings, file_secret_settings


class StationSettings(BaseSettings):
    """Base Settings with recursive construct method"""

    _fixable_fields: list[str] = []

    @classmethod
    def construct(cls, _fields_set=None, **values):
        """Construct a new model object from the function arguments. Recursively construct nested models.
        If an environment variable is available for a field, it will override the value passed in the function
        arguments.

        Args:
            _fields_set: Optional set of fields for hte model
            **values: The values to set for the model

        Returns:
            A new model object
        """
        m = cls.__new__(cls)
        fields_values = {}

        config = cls.__config__

        for name, field in cls.__fields__.items():
            key = field.alias
            if (
                key not in values and config.allow_population_by_field_name
            ):  # Added this to allow population by field name
                key = name

            if key in values:
                if (
                    values[key] is None and not field.required
                ):  # Moved this check since None value can be passed for Optional nested field
                    fields_values[name] = field.get_default()
                else:
                    # check for union type
                    if type(field.type_) == type(Union):
                        print("Found union type")

                    if issubclass(field.type_, BaseSettings):
                        # check if the field is a list of models
                        if field.shape == 2:
                            fields_values[name] = [
                                field.type_.construct(**e) for e in values[key]
                            ]
                        else:
                            fields_values[name] = field.outer_type_.construct(
                                **values[key]
                            )
                    else:
                        fields_values[name] = values[key]
            elif not field.required:
                fields_values[name] = field.get_default()

        # override with env variables
        # read the environment variables specs of the models
        # override the fields only if they are not of object type
        for name, field in m.__fields__.items():
            if issubclass(field.type_, BaseSettings):
                pass
            else:
                env_name = f"{config.env_prefix}{name.upper()}"
                env = os.getenv(env_name)
                if env:
                    # try to convert the value to the type of the field
                    if field.type_ == bool:
                        env = env.lower() == "true"
                    elif field.type_ == int:
                        env = int(env)
                    elif field.type_ == float:
                        env = float(env)
                    # set the value
                    fields_values[name] = env

        object.__setattr__(m, "__dict__", fields_values)
        if _fields_set is None:
            _fields_set = set(values.keys())
        object.__setattr__(m, "__fields_set__", _fields_set)

        m._init_private_attributes()
        return m

    def get_fix(self, field: str) -> ConfigItemFix:
        """Returns a fix for the field.
        Returns:
            ConfigItemFix: A fix for the field or None.
        """

        if field not in self.__class__.__fields__:
            raise ValueError(
                f"Field {field} does not exist in model {self.__class__.__name__}"
            )


class ServiceSettings(StationSettings):
    """Model that contains common settings for configuring the connection to a service
    as admin user.
    """

    admin_user: str
    admin_password: SecretStr
    host: str
    port: Optional[int] = None
    # validator for admin password
    _admin_password = admin_validator()

    _fixable_fields = ["admin_password"]

    def get_fix(self, field: str) -> ConfigItemFix | None:
        """Returns a dict with the settings that can be used to fix the service.
        Returns:
            dict: dict with the settings that can be used to fix the service.
        """
        super().get_fix(field)
        match field:
            case "admin_password":
                if hasattr(self, "admin_password"):
                    if isinstance(self.admin_password, SecretStr):
                        val = self.admin_password.get_secret_value()
                    else:
                        val = self.admin_password
                else:
                    val = None
                return ConfigItemFix.admin_password(value=val)

            case _:
                return None


class CentralSettings(StationSettings):
    """Model that contains settings for configuring the connection to the central API as robot user."""

    url: AnyHttpUrl
    robot_id: str
    robot_secret: SecretStr
    private_key: str
    private_key_password: Optional[SecretStr] = None

    _private_key_validator = file_readable_validator("private_key")

    Config = StationSettingsConfig.with_prefix("STATION_CENTRAL_")

    def get_fix(self, field: str) -> ConfigItemFix | None:
        """Returns a dict with the settings that can be used to fix the service.
        Returns:
            dict: dict with the settings that can be used to fix the service.
        """
        super().get_fix(field)
        match field:
            case "private_key":
                val = None
                if hasattr(self, "private_key"):
                    val = self.private_key
                return ConfigItemFix.private_key(value=val)

            case _:
                return None


class HttpSettings(StationSettings):
    """General HTTP settings for the station."""

    port: Optional[int] = 80

    Config = StationSettingsConfig.with_prefix("STATION_HTTP_")


class Certificate(StationSettings):
    """Model that contains settings for configuring the HTTPS certificates."""

    cert: str
    key: str

    _cert_validator = file_readable_validator("cert")
    _key_validator = file_readable_validator("key")

    Config = StationSettingsConfig.with_prefix("STATION_HTTPS_CERTIFICATES_")


class HttpsSettings(StationSettings):
    """MOdel that contains the stations HTTPS settings."""

    port: Optional[int] = 443
    domain: str
    certificate: Optional[Certificate] = None

    Config = StationSettingsConfig.with_prefix("STATION_HTTPS_")


class TraefikSettings(StationSettings):
    """Configuration options for the Stations Traefik instance."""

    dashboard: Optional[bool] = False
    dashboard_port: Optional[int] = 8081

    Config = StationSettingsConfig.with_prefix("STATION_TRAEFIK_")


class RegistrySettings(StationSettings):
    """Model that contains settings for configuring the connection to the central container registry."""

    address: str
    user: str
    password: SecretStr
    project: str

    _password = admin_validator("password")

    Config = StationSettingsConfig.with_prefix("STATION_REGISTRY_")


class DBSettings(ServiceSettings):
    """Model that contains settings for configuring the connection to the database."""

    Config = StationSettingsConfig.with_prefix("STATION_DB_")


class AirflowSettings(ServiceSettings):
    """Model that contains settings for configuring the connection to the airflow instance.
    With the option of provididing a custom config and extra DAGs directory.
    """

    config_path: str | None = None
    extra_dags_dir: str | None = None

    @validator("config_path")
    def validate_config_path(cls, value: str | None) -> str | None:
        if value is None:
            return value
        return validate_file_readable(value)

    @validator("extra_dags_dir")
    def validate_extra_dags_dir(cls, value: str | None) -> str | None:
        if value is None:
            return value

        if os.path.isdir(value):
            return value

    Config = StationSettingsConfig.with_prefix("STATION_AIRFLOW_")


class MinioSettings(ServiceSettings):
    """Model that contains settings for configuring the stations minio instance."""

    Config = StationSettingsConfig.with_prefix("STATION_MINIO_")


class AuthSettings(StationSettings):
    """Model that contains settings for configuring the stations authup instance."""

    host: str
    port: Optional[int] = 3001
    admin_user: Optional[str] = "admin"
    robot_id: Optional[str]
    robot_secret: Optional[SecretStr]

    Config = StationSettingsConfig.with_prefix("STATION_AUTH_")


class APISettings(StationSettings):
    """Model that contains settings for configuring the station API."""

    port: Optional[int] = 8000
    fernet_key: SecretStr
    database: Optional[str] = "pht_station"

    Config = StationSettingsConfig.with_prefix("STATION_API_")

    # todo validate fernet key

    @validator("fernet_key")
    def validate_fernet_key(cls, value: SecretStr) -> SecretStr:
        return validate_fernet_key(value)

    def get_fix(self, field: str) -> ConfigItemFix | None:
        """Returns a dict with the settings that can be used to fix the service.
        Returns:
            dict: dict with the settings that can be used to fix the service.
        """
        super().get_fix(field)
        match field:
            case "fernet_key":
                return ConfigItemFix.fernet_key(value=self.fernet_key)
            case _:
                return None


class RedisSettings(StationSettings):
    """Model that contains settings for configuring the stations redis instance."""

    host: str
    port: Optional[int] = 6379
    admin_password: Optional[SecretStr] = None
    database: Optional[int] = 0

    _admin_password = admin_validator()

    Config = StationSettingsConfig.with_prefix("STATION_REDIS_")

    def get_fix(self, field: str) -> ConfigItemFix:
        super().get_fix(field)

        match field:
            case "admin_password":
                return ConfigItemFix.admin_password(value=self.admin_password)
            case _:
                return None


class StationConfig(StationSettings):
    """Unified model that contains all settings for configuring the station."""

    id: str
    environment: ApplicationEnvironment
    admin_password: SecretStr
    data_dir: str
    version: str
    central: CentralSettings
    http: HttpSettings
    https: HttpsSettings
    traefik: TraefikSettings
    registry: RegistrySettings
    api: APISettings
    db: DBSettings
    airflow: AirflowSettings
    minio: MinioSettings
    auth: AuthSettings
    redis: RedisSettings

    _admin_password = admin_validator()
    _data_dir = dir_validator("data_dir")

    def display(self):
        pprint(self, expand_all=True)

    @classmethod
    def from_file(cls, path: str) -> "StationConfig":
        with open(path, "r") as f:
            # load yaml as dict
            return cls.parse_obj(yaml.safe_load(f))

    def get_fix(self, field: str) -> ConfigItemFix | None:
        """Returns a dict with the settings that can be used to fix the service.
        Returns:
            dict: dict with the settings that can be used to fix the service.
        """
        super().get_fix(field)
        match field:
            case "admin_password":
                val = None
                if hasattr(self, "admin_password"):
                    val = self.admin_password

                return ConfigItemFix.admin_password(value=val)

            case "data_dir":
                return ConfigItemFix.no_fix(
                    value=self.data_dir,
                    suggestion="Check that the path is correct and that the user has access to it.",
                )

            case "id":
                return ConfigItemFix.no_fix(
                    value=self.id,
                    suggestion="Enter the station id that is displayed in the central PHT user interface.",
                )

            case _:
                print(f"no fix for {field}")
                return None

    Config = StationSettingsConfig.with_prefix("STATION_")



from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from station.app.models.discovery import DataSetSummary
from station.app.schemas.discovery import SummaryCreate, SummaryUpdate

from .base import CreateSchemaType, CRUDBase, ModelType, Optional


class CRUDDiscoveries(CRUDBase[DataSetSummary, SummaryCreate, SummaryUpdate]):
    def create(self, db: Session, *, obj_in: CreateSchemaType) -> Optional[ModelType]:
        obj_in_data = jsonable_encoder(obj_in)
        db_obj = self.model(**obj_in_data)
        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)

        return db_obj


discoveries = CRUDDiscoveries(DataSetSummary)



from enum import Enum
from typing import List

import requests
from fastapi import Depends, HTTPException
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from loguru import logger
from requests import HTTPError

from station.app.cache import redis_cache
from station.app.config import settings
from station.app.schemas.users import User, UserPermission


class TokenCacheKeys(str, Enum):
    robot_token = "robot-token"
    user_token_prefix = "user-token-"


def get_robot_token(
    robot_id: str = None, robot_secret: str = None, token_url: str = None
) -> str:
    """
    Get robot token from auth server.
    """
    # todo token caching

    logger.debug("Getting robot token")

    if not token_url:
        token_url = settings.config.auth.token_url
    if not robot_id:
        robot_id = settings.config.auth.robot_id
    if not robot_secret:
        robot_secret = settings.config.auth.robot_secret.get_secret_value()

    # try to read the token from cache and return it if it exists
    cached_token = redis_cache.get(TokenCacheKeys.robot_token.value)
    if cached_token:
        logger.debug("Found cached robot token")
        return cached_token
    else:
        # get a new token from the auth server
        logger.debug(f"Requesting new robot token from {token_url}")
        data = {
            "id": robot_id,
            "secret": robot_secret,
            "grant_type": "robot_credentials",
        }

        response = requests.post(token_url, data=data).json()

        # parse values from response and set cache
        token = response.get("access_token")
        ttl = response.get("expires_in")

        redis_cache.set(TokenCacheKeys.robot_token.value, token, ttl)

        return token


def validate_user_token(token: str, user_url: str = None) -> User:
    """
    Validate a user token against the auth server and parse a user object from the response.
    Args:
        token: token to validate
        user_url: user url of the auth server

    Returns:
        User object parsed from the auth server response
    """
    # todo token caching
    if user_url is None:
        user_url = settings.config.auth.user_url
    url = f"{user_url}/@me"
    logger.debug(f"Validating user token against {url}")
    headers = {"Authorization": f"Bearer {token}"}
    r = requests.get(url, headers=headers)
    r.raise_for_status()
    user = User(**r.json())
    return user


def get_current_user(token: str, token_url: str = None) -> User:
    """
    Validate a user token against the auth server and parse a user object from the response.
    Args:
        token: token to validate
        robot_token: the robot token to request token validation
        token_url: token url of the auth server

    Returns:
        User object parsed from the auth server response
    """

    logger.debug("Validating bearer token")
    if not settings.is_initialized:
        logger.error("Found uninitialized setting.... Initializing settings")
        settings.setup()
    if token_url is None:
        token_url = settings.config.auth.token_url
    try:
        user = validate_user_token(token=token)

        return user
    except HTTPError as e:
        logger.error(f"Error validating user token: {e}")
        if e.response.status_code == 401:
            raise HTTPException(status_code=401, detail="Invalid token")
        elif e.response.status_code == 400:
            # attempt refresh robot token
            try:

                get_robot_token(
                    robot_id=settings.config.auth.robot_id,
                    robot_secret=settings.config.auth.robot_secret.get_secret_value(),
                    token_url=token_url,
                )

                user = validate_user_token(token=token)
                return user
            except HTTPError:
                raise HTTPException(status_code=401, detail="Invalid token")


def get_user_permissions(user: User, token_url: str = None) -> List[UserPermission]:
    # todo token caching
    logger.debug(f"Getting user permissions for {user.name}")
    if token_url is None:
        token_url = settings.config.auth.token_url
    url = f"{token_url}/introspect"
    headers = {"Authorization": f"Bearer {user.token}"}
    r = requests.get(url, headers=headers)
    r.raise_for_status()
    permissions = [UserPermission(**p) for p in r.json().get("permissions")]
    return permissions


# def authorized_user(token: HTTPAuthorizationCredentials = Depends(HTTPBearer()),
#                     permissions: List[UserPermission] = None) -> User:
def authorized_user(token: HTTPAuthorizationCredentials = Depends(HTTPBearer())):
    """
    Authorize a user by checking if they are in the allowed users list.
    Args:
        token: Bearer token from http header
        permissions: list of permissions to check

    Returns:
        User object if authorized

    """
    logger.debug("Authorizing user")
    user = get_current_user(token=token.credentials)
    # if permissions:
    #     user.permissions = get_user_permissions(user)
    #     # todo validate permissions

    return user



from typing import List, Optional

from pydantic import BaseModel


class StationKeys(BaseModel):
    station_id: int
    signing_key: str
    sharing_key: str
    key_signature: Optional[str] = None

    class Config:
        orm_mode = True


class BroadCastKeysSchema(BaseModel):
    train_id: int
    iteration: int
    keys: List[StationKeys]


class GetCyphersRequest(BaseModel):
    station_id: int
    iteration: int



import os
import pprint

import pytest
from dotenv import find_dotenv, load_dotenv

from station.common.clients.central.central_client import CentralApiClient


@pytest.fixture
def central_client():
    load_dotenv(find_dotenv())
    url = os.getenv("CENTRAL_API_URL")
    client_id = os.getenv("STATION_ROBOT_ID")
    client_secret = os.getenv("STATION_ROBOT_SECRET")
    return CentralApiClient(url, client_id, client_secret)


def test_get_token(central_client):
    token = central_client._get_token()

    assert token is not None


def test_get_registry_credentials(central_client):
    registry_credentials = central_client.get_registry_credentials(
        os.getenv("STATION_ID")
    )
    pprint.pprint(registry_credentials)
    assert registry_credentials


def test_update_public_key(central_client):
    public_key = "test_public_key".encode("utf-8").hex()
    station_id = os.getenv("STATION_ID")
    response = central_client.update_public_key(station_id, public_key)
    print(response)


def test_get_trains_for_station(central_client):
    station_id = os.getenv("STATION_ID")
    response = central_client.get_trains(station_id)
    assert response
    print(response)



import click

from station.ctl.fhir import manage
from station.ctl.fhir.setup import setup_cli


@click.command(help="Manage FHIR servers associated with the station.")
@click.option(
    "-c",
    "--command",
    help="The command to execute",
    type=click.Choice(["setup", "start", "stop", "list", "remove"]),
    required=True,
)
@click.option("-n", "--name", help="The name of the server to run the command against")
@click.pass_context
def fhir(ctx, command, name):
    if command == "setup":
        setup_cli(ctx, name)
    elif command == "start":
        manage.start_server(name)
    elif command == "stop":
        manage.stop_server(name)
    elif command == "remove":
        manage.remove_server(name)
    elif command == "list":
        click.echo("list")



import os

import click

from station.common.constants import Icons
from station.ctl.config import find_config, load_config
from station.ctl.config.command import settings
from station.ctl.fhir.command import fhir
from station.ctl.install.command import install


@click.group()
@click.option(
    "--config",
    type=click.Path(exists=True),
    help="Path to config file. If none is given assumes config file is located in current working directory.",
)
@click.pass_context
def cli(ctx, config):
    if config:
        ctx.obj["station_config"] = load_config(config)
    else:
        click.echo(
            f"No config file given. Looking for config file in current directory({str(os.getcwd())})... ",
            nl=False,
        )
        try:
            config_dict, path = find_config(os.getcwd())
            if not ctx.obj:
                ctx.obj = {}
            ctx.obj["station_config"] = config_dict
            ctx.obj["config_path"] = path
            click.echo(Icons.CHECKMARK.value)
        except FileNotFoundError:
            click.echo(Icons.CROSS.value)
            click.echo("No config file found.")


@cli.command(help="Uninstall the station software. Coming soon..")
@click.pass_context
def uninstall(ctx):
    # todo
    print(ctx.obj)
    click.echo("Coming soon...")


@cli.command(help="Update the station software. Coming soon..")
@click.pass_context
def update(ctx):
    # todo
    click.echo("Coming soon...")


cli.add_command(install)
cli.add_command(settings)
cli.add_command(fhir)

if __name__ == "__main__":
    cli()



import uuid
from datetime import datetime

from sqlalchemy import JSON, Column, DateTime, ForeignKey, String
from sqlalchemy.dialects.postgresql import UUID

from station.app.db.base_class import Base


class DataSet(Base):
    __tablename__ = "datasets"
    id = Column(UUID(as_uuid=True), primary_key=True, index=True, default=uuid.uuid4)
    created_at = Column(DateTime, default=datetime.now())
    updated_at = Column(DateTime, nullable=True)
    proposal_id = Column(String, nullable=True)
    name = Column(String)
    data_type = Column(String, nullable=True)
    storage_type = Column(String, nullable=True)
    access_path = Column(String, nullable=True)
    fhir_server = Column(UUID, ForeignKey("fhir_servers.id"), nullable=True)
    summary = Column(JSON, nullable=True)



import os

from dotenv import find_dotenv, load_dotenv
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from station.app.db.base import Base

# Create new sqlite database for testing

# load the .env file
load_dotenv(find_dotenv())

SQLALCHEMY_DATABASE_URL = os.getenv("TEST_DB", "sqlite:///./test.db")

if SQLALCHEMY_DATABASE_URL.startswith("sqlite"):
    engine = create_engine(
        SQLALCHEMY_DATABASE_URL, connect_args={"check_same_thread": False}
    )
else:
    engine = create_engine(SQLALCHEMY_DATABASE_URL)

TestingSessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

Base.metadata.drop_all(bind=engine)
Base.metadata.create_all(bind=engine)


def override_get_db():
    try:
        db = TestingSessionLocal()
        yield db
    finally:
        db.close()



import uuid
from datetime import datetime

from sqlalchemy import JSON, Boolean, Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import relationship

from station.app.db.base_class import Base


class LocalTrainState(Base):
    __tablename__ = "local_train_states"
    id = Column(Integer, primary_key=True, index=True)
    train_id = Column(UUID(as_uuid=True), ForeignKey("local_trains.id"))
    last_execution = Column(DateTime, nullable=True)
    num_executions = Column(Integer, default=0)
    status = Column(String, default="inactive")
    configuration_state = Column(String, default="initialized")


class LocalTrainExecution(Base):
    __tablename__ = "local_train_executions"
    id = Column(UUID(as_uuid=True), primary_key=True, index=True, default=uuid.uuid4)
    train_id = Column(UUID, ForeignKey("local_trains.id"))
    airflow_dag_run = Column(String, nullable=True, unique=True)
    config_id = Column(Integer, ForeignKey("docker_train_configs.id"), nullable=True)
    dataset_id = Column(UUID(as_uuid=True), ForeignKey("datasets.id"), nullable=True)
    start = Column(DateTime, default=datetime.now())
    finish = Column(DateTime, nullable=True)


class LocalTrainMasterImage(Base):
    __tablename__ = "local_train_master_images"
    id = Column(UUID(as_uuid=True), primary_key=True, index=True, default=uuid.uuid4)
    registry = Column(String, nullable=True)
    group = Column(String, nullable=True)
    artifact = Column(String, nullable=True)
    tag = Column(String, default="latest")
    image_id = Column(String, nullable=True, unique=True)
    created_at = Column(DateTime, default=datetime.now())
    updated_at = Column(DateTime, nullable=True)


class LocalTrain(Base):
    __tablename__ = "local_trains"
    id = Column(UUID(as_uuid=True), primary_key=True, index=True, default=uuid.uuid4)
    name = Column(String, nullable=True)
    master_image_id = Column(
        UUID(as_uuid=True), ForeignKey("local_train_master_images.id"), nullable=True
    )
    entrypoint = Column(String, nullable=True)
    custom_image = Column(String, nullable=True)
    command = Column(String, nullable=True)
    command_args = Column(String, nullable=True)
    fhir_query = Column(String, nullable=True)
    created_at = Column(DateTime, default=datetime.now())
    updated_at = Column(DateTime, nullable=True)
    state = relationship("LocalTrainState", cascade="all,delete", uselist=False)
    dataset_id = Column(UUID(as_uuid=True), ForeignKey("datasets.id"), nullable=True)
    config_id = Column(
        UUID(as_uuid=True), ForeignKey("local_train_configs.id"), nullable=True
    )


class LocalTrainConfig(Base):
    __tablename__ = "local_train_configs"
    id = Column(UUID(as_uuid=True), primary_key=True, index=True, default=uuid.uuid4)
    name = Column(String, unique=True)
    created_at = Column(DateTime, default=datetime.now())
    updated_at = Column(DateTime, nullable=True)
    airflow_config_json = Column(JSON, nullable=True)
    trains = relationship("LocalTrain")
    cpu_requirements = Column(JSON, nullable=True)
    gpu_requirements = Column(JSON, nullable=True)
    auto_execute = Column(Boolean, default=False)



from typing import List

from sqlalchemy.orm import Session

from station.app.models.notification import Notification
from station.app.schemas.notifications import NotificationCreate, NotificationUpdate

from .base import CRUDBase


class CRUDNotifications(CRUDBase[Notification, NotificationCreate, NotificationUpdate]):
    def read_notifications_for_user(self, db: Session, user: str) -> List[Notification]:
        return db.query(Notification).filter(Notification.target_user == user).all()


notifications = CRUDNotifications(Notification)



import os

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

if os.getenv("STATION_DB"):
    SQLALCHEMY_DATABASE_URL = os.getenv("STATION_DB")
else:
    SQLALCHEMY_DATABASE_URL = "postgresql+psycopg2://admin:admin@localhost/pht_station"


engine = create_engine(
    SQLALCHEMY_DATABASE_URL,  # connect_args={"check_same_thread": False}  For sqlite db
)

SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)



import re
from typing import List

import click

import docker
from station.common.constants import DockerNetworks, Icons, ServiceImages


def setup_cli(ctx: click.Context, name: str):
    if not name:
        name = click.prompt("Please enter a name for the server")

    name = _validate_server_name(name)
    port = click.prompt(
        "If you would like to expose the server on the host machine please enter the port number",
        default="",
    )
    domain = ctx.obj["https"]["domain"]
    click.echo(
        f"Setting up server with base url: {domain}/fhir-servers/{name}... ", nl=False
    )
    try:
        setup_server(domain, name, port)
        click.echo(f"{Icons.CHECKMARK.value}")
    except Exception as e:
        click.echo(Icons.CROSS.value)
        click.echo(f"Error: {e}")


def setup_server(domain: str, name: str, port: str = None):
    client = docker.from_env()
    volume = _setup_volume(client, name)
    labels = _make_labels(domain, name)
    env = _make_env(domain, name)
    client.containers.run(
        ServiceImages.BLAZE.value,
        volumes=volume,
        labels=labels,
        environment=env,
        detach=True,
        name=name,
        network=DockerNetworks.STATION.value,
        ports={port: "8080"} if port else None,
    )


def _setup_volume(client: docker.DockerClient, name: str) -> List[str]:
    volume_name = f"blaze-{name}"
    client.volumes.create(name=volume_name)
    volume_list = [f"{volume_name}:/app/data"]
    return volume_list


def _make_labels(domain: str, name: str) -> dict:
    return {
        "traefik.enable": "true",
        f"traefik.http.routers.{'fhir-' + name}.tls": "true",
        f'traefik.http.routers.{"fhir-" + name}.rule': f'Host("{domain}") && PathPrefix("/fhir-servers/{name}")',
        f"traefik.http.services.{'fhir-' + name}.loadbalancer.server.port": "8080",
    }


def _make_env(domain: str, name: str) -> List[str]:
    # todo add station auth server oidc connection variables
    return [
        "JAVA_TOOL_OPTIONS=-Xmx2g",
        f"BASE_URL=https://{domain}/fhir-servers/{name}",
    ]


def _validate_server_name(name: str) -> str:
    match = re.match(r"^[a-zA-Z\d_-]*$", name)
    if not match:
        name = click.prompt(
            f"Invalid server name: {name}. Enter a valid server name containing only [a-zA-Z0-9_-]."
        )
        return _validate_server_name(name)
    return name



import os

from station.app.schemas.datasets import DataSet
from station.app.schemas.local_trains import LocalTrain
from station.app.schemas.trains import Train
from station.common.clients.base import BaseClient
from station.common.clients.resource_client import ResourceClient
from station.common.clients.station.local_trains import LocalTrainClient


class StationAPIClient(BaseClient):
    local_trains: LocalTrainClient
    datasets: ResourceClient
    trains: ResourceClient

    def __init__(
        self,
        base_url: str,
        auth_url: str = None,
        username: str = None,
        password: str = None,
    ):
        super().__init__(
            base_url=base_url, auth_url=auth_url, username=username, password=password
        )

        self.local_trains = LocalTrainClient(base_url, "local-trains", LocalTrain, self)
        self.datasets = ResourceClient(base_url, "datasets", DataSet, client=self)
        self.trains = ResourceClient(base_url, "trains/docker", Train, client=self)

    @classmethod
    def from_env(cls):
        base_url = os.getenv("STATION_API_URL")
        username = os.getenv("STATION_USER")
        password = os.getenv("STATION_PASSWORD")
        auth_url = os.getenv("AUTH_URL")

        return cls(
            base_url=base_url, username=username, password=password, auth_url=auth_url
        )



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel


class HealthStatus(str, Enum):
    healthy = "healthy"
    stopped = "stopped"
    error = "error"


class StatusDocker(BaseModel):
    Id: str


class ServiceStatus(BaseModel):
    name: str
    status: HealthStatus


class DiskUsage(BaseModel):
    total: int
    used: int
    free: int
    percent: float


class MemoryUsage(BaseModel):
    total: int
    available: int
    free: int
    used: int
    percent: float


class HardwareResources(BaseModel):
    cpu: List[float]
    memory: MemoryUsage
    disk: DiskUsage
    gpu: Optional[float] = None


class StationStatus(BaseModel):
    services: List[ServiceStatus]
    hardware: HardwareResources
    docker: Optional[dict] = None



from fastapi import APIRouter, Depends
from sqlalchemy.orm import Session

from station.app.api import dependencies
from station.app.schemas.users import User

router = APIRouter()


@router.get("/config")
def get_station_config(db: Session = Depends(dependencies.get_db)):
    # TODO store station configuration either inside yaml/config file or in db and read it here
    pass


@router.get("/config/test")
def test_station_config():
    pass


@router.put("/config")
def update_station_config():
    # TODO allow for updates and storage of configuration values for a station
    pass



from dotenv import find_dotenv, load_dotenv
from fastapi.testclient import TestClient

from station.app.api.dependencies import get_db
from station.app.main import app

from .test_db import override_get_db

load_dotenv(find_dotenv())

app.dependency_overrides[get_db] = override_get_db

client = TestClient(app)


def test_station_status():
    resp = client.get("api/station/status")
    assert resp.status_code == 200



from typing import List, Optional, Union

from pydantic import BaseModel


class NumericalData(BaseModel):
    attribute_name: Optional[str]
    mean: Optional[float]
    min: Optional[float]
    max: Optional[float]


class CategoricalCount(BaseModel):
    category_value: Optional[str]
    count: Optional[int]


class CategoricalData(BaseModel):
    attribute_name: Optional[str]
    value_counts: Optional[List[CategoricalCount]]


class CategoricalDataLocal(CategoricalData):
    most_frequent_element: Optional[Union[str, int]]
    frequency: Optional[int]


class StructuredData(BaseModel):
    data_summary: Optional[List[Union[NumericalData, CategoricalData]]]


class UnstructuredData(BaseModel):
    target_counts: Optional[CategoricalCount]
    mean_size: Optional[float]


class DataSetSummary(BaseModel):
    proposal_id: Optional[int]
    count: Optional[int]
    information: Optional[Union[StructuredData, UnstructuredData]]


class SummaryCreate(DataSetSummary):
    pass


class SummaryUpdate(DataSetSummary):
    pass



import datetime
from typing import List, Optional

from pydantic import BaseModel


def to_camel(string: str) -> str:
    split = string.split("_")
    return split[0] + "".join(word.capitalize() for word in split[1:])


class UserPermission(BaseModel):
    id: str
    negation: bool
    power: int
    condition: Optional[str] = None
    fields: Optional[List[str]] = None


class User(BaseModel):
    id: str
    name: str
    email: Optional[str] = None
    active: bool
    token: Optional[str] = None
    created_at: datetime.datetime
    updated_at: Optional[datetime.datetime] = None
    realm_id: str
    display_name: str
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    name_locked: bool
    permissions: Optional[List[UserPermission]] = None
    status: Optional[str] = None
    status_message: Optional[str] = None


class Token(BaseModel):
    exp: int
    iat: int
    iss: str
    remote_address: str
    sub: str
    sub_kind: str

    class Config:
        alias_generator = to_camel


class UserResponse(BaseModel):
    user: User
    token: dict



from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

import pytorch_lightning as pl
from pytorch_lightning import Callback, LightningDataModule, LightningModule
from pytorch_lightning.accelerators import Accelerator
from pytorch_lightning.loggers import LightningLoggerBase
from pytorch_lightning.profiler import BaseProfiler
from torch.utils.data import DataLoader
from torchvision.transforms import Compose, ToTensor

from station.worker.loader import BaseLoader


class FederatedTrainer(pl.Trainer):
    def __init__(
        self,
        logger: Union[LightningLoggerBase, Iterable[LightningLoggerBase], bool] = True,
        checkpoint_callback: bool = True,
        callbacks: Optional[List[Callback]] = None,
        default_root_dir: Optional[str] = None,
        gradient_clip_val: float = 0,
        process_position: int = 0,
        num_nodes: int = 1,
        num_processes: int = 1,
        gpus: Optional[Union[List[int], str, int]] = None,
        auto_select_gpus: bool = False,
        tpu_cores: Optional[Union[List[int], str, int]] = None,
        log_gpu_memory: Optional[str] = None,
        progress_bar_refresh_rate: int = 1,
        overfit_batches: Union[int, float] = 0.0,
        track_grad_norm: Union[int, float, str] = -1,
        check_val_every_n_epoch: int = 1,
        fast_dev_run: Union[int, bool] = False,
        accumulate_grad_batches: Union[int, Dict[int, int], List[list]] = 1,
        max_epochs: int = 1000,
        min_epochs: int = 1,
        max_steps: Optional[int] = None,
        min_steps: Optional[int] = None,
        limit_train_batches: Union[int, float] = 1.0,
        limit_val_batches: Union[int, float] = 1.0,
        limit_test_batches: Union[int, float] = 1.0,
        val_check_interval: Union[int, float] = 1.0,
        flush_logs_every_n_steps: int = 100,
        log_every_n_steps: int = 50,
        accelerator: Optional[Union[str, Accelerator]] = None,
        sync_batchnorm: bool = False,
        precision: int = 32,
        weights_summary: Optional[str] = "top",
        weights_save_path: Optional[str] = None,
        num_sanity_val_steps: int = 2,
        truncated_bptt_steps: Optional[int] = None,
        resume_from_checkpoint: Optional[Union[Path, str]] = None,
        profiler: Optional[Union[BaseProfiler, bool, str]] = None,
        benchmark: bool = False,
        deterministic: bool = False,
        reload_dataloaders_every_epoch: bool = False,
        auto_lr_find: Union[bool, str] = False,
        replace_sampler_ddp: bool = True,
        terminate_on_nan: bool = False,
        auto_scale_batch_size: Union[str, bool] = False,
        prepare_data_per_node: bool = True,
        plugins: Optional[Union[str, list]] = None,
        amp_backend: str = "native",
        amp_level: str = "O2",
        distributed_backend: Optional[str] = None,
        automatic_optimization: Optional[bool] = None,
        move_metrics_to_cpu: bool = False,
        enable_pl_optimizer: bool = None,
    ):
        super().__init__(
            logger,
            checkpoint_callback,
            callbacks,
            default_root_dir,
            gradient_clip_val,
            process_position,
            num_nodes,
            num_processes,
            gpus,
            auto_select_gpus,
            tpu_cores,
            log_gpu_memory,
            progress_bar_refresh_rate,
            overfit_batches,
            track_grad_norm,
            check_val_every_n_epoch,
            fast_dev_run,
            accumulate_grad_batches,
            max_epochs,
            min_epochs,
            max_steps,
            min_steps,
            limit_train_batches,
            limit_val_batches,
            limit_test_batches,
            val_check_interval,
            flush_logs_every_n_steps,
            log_every_n_steps,
            accelerator,
            sync_batchnorm,
            precision,
            weights_summary,
            weights_save_path,
            num_sanity_val_steps,
            truncated_bptt_steps,
            resume_from_checkpoint,
            profiler,
            benchmark,
            deterministic,
            reload_dataloaders_every_epoch,
            auto_lr_find,
            replace_sampler_ddp,
            terminate_on_nan,
            auto_scale_batch_size,
            prepare_data_per_node,
            plugins,
            amp_backend,
            amp_level,
            distributed_backend,
            automatic_optimization,
            move_metrics_to_cpu,
            enable_pl_optimizer,
        )

    def setup_trainer(self, model: LightningModule):
        super().setup_trainer(model)

    def fit(
        self,
        model: LightningModule,
        train_dataloader: Optional[DataLoader] = None,
        val_dataloaders: Optional[Union[DataLoader, List[DataLoader]]] = None,
        datamodule: Optional[LightningDataModule] = None,
    ):
        return super().fit(model, train_dataloader, val_dataloaders, datamodule)

    def train(self):
        super().train()

    def run_evaluation(self, max_batches=None):
        return super().run_evaluation(max_batches)

    def track_output_for_epoch_end(self, outputs, output):
        return super().track_output_for_epoch_end(outputs, output)

    def run_test(self):
        return super().run_test()

    def run_sanity_check(self, ref_model):
        super().run_sanity_check(ref_model)

    def test(
        self,
        model: Optional[LightningModule] = None,
        test_dataloaders: Optional[Union[DataLoader, List[DataLoader]]] = None,
        ckpt_path: Optional[str] = "best",
        verbose: bool = True,
        datamodule: Optional[LightningDataModule] = None,
    ):
        return super().test(model, test_dataloaders, ckpt_path, verbose, datamodule)

    def tune(
        self,
        model: LightningModule,
        train_dataloader: Optional[DataLoader] = None,
        val_dataloaders: Optional[Union[DataLoader, List[DataLoader]]] = None,
        datamodule: Optional[LightningDataModule] = None,
    ):
        super().tune(model, train_dataloader, val_dataloaders, datamodule)

    def call_setup_hook(self, model):
        super().call_setup_hook(model)

    def _reset_result_and_set_hook_fx_name(self, hook_name):
        return super()._reset_result_and_set_hook_fx_name(hook_name)

    def _cache_logged_metrics(self):
        super()._cache_logged_metrics()

    def call_hook(self, hook_name, *args, **kwargs):
        return super().call_hook(hook_name, *args, **kwargs)

    def on_batch_end(self):
        print("Batch Ended")
        super().on_batch_end()

    @staticmethod
    def available_plugins():
        return super().available_plugins()


if __name__ == "__main__":
    transform = Compose([ToTensor()])

    loader = BaseLoader(
        data_set="cifar/batch_1",
        transform=transform,
        config={"batch_size": 6, "num_workers": 0},
    ).make_data_loader()

    trainer = FederatedTrainer()

    trainer.fit()



from cryptography.fernet import Fernet

from station.ctl.config import validators


def test_validate_db_config():
    # invalid password
    db_config = {"admin_user": "admin", "admin_password": "admin"}

    result = validators.validate_db_config(db_config)
    issues = [
        issue
        for issue in result
        if issue.status != validators.ConfigItemValidationStatus.VALID
    ]
    assert len(issues) == 1
    assert issues[0].display_field == "db.admin_password"

    # invalid username and password
    db_config = {"admin_user": "", "admin_password": "admin"}

    result = validators.validate_db_config(db_config)
    issues = [
        issue
        for issue in result
        if issue.status != validators.ConfigItemValidationStatus.VALID
    ]
    assert len(issues) == 2

    # valid config
    db_config = {"admin_user": "admin", "admin_password": "valid_password"}

    result = validators.validate_db_config(db_config)
    issues = [
        issue
        for issue in result
        if issue.status != validators.ConfigItemValidationStatus.VALID
    ]
    assert len(issues) == 0

    # no config given
    db_config = None

    result = validators.validate_db_config(db_config)
    issues = [
        issue
        for issue in result
        if issue.status != validators.ConfigItemValidationStatus.VALID
    ]
    assert len(issues) == 1


def test_validate_api_config():
    api_config = {"fernet_key": "your_fernet_key"}

    result = validators.validate_api_config(api_config)
    issues = [
        issue
        for issue in result
        if issue.status != validators.ConfigItemValidationStatus.VALID
    ]
    assert len(issues) == 1
    assert issues[0].display_field == "api.fernet_key"

    api_config = {"fernet_key": "dsa"}

    result = validators.validate_api_config(api_config)
    issues = [
        issue
        for issue in result
        if issue.status != validators.ConfigItemValidationStatus.VALID
    ]
    assert len(issues) == 1
    assert issues[0].display_field == "api.fernet_key"

    api_config = {"fernet_key": Fernet.generate_key().decode()}
    result = validators.validate_api_config(api_config)
    issues = [
        issue
        for issue in result
        if issue.status != validators.ConfigItemValidationStatus.VALID
    ]
    assert len(issues) == 0


def test_validate_minio_config():
    pass


def test_validate_airflow_config():
    pass


def test_validate_central_config():
    pass



from sqlalchemy.orm import Session

from station.app.config import settings
from station.app.crud.base import CRUDBase
from station.app.models.local_trains import LocalTrainMasterImage
from station.app.schemas import local_trains as schemas
from station.common.clients.harbor_client import HarborClient


class CRUDLocalTrainMasterImage(
    CRUDBase[
        LocalTrainMasterImage,
        schemas.LocalTrainMasterImageCreate,
        schemas.LocalTrainMasterImageUpdate,
    ]
):
    def get_by_image_id(self, db: Session, image_id: str):
        return (
            db.query(LocalTrainMasterImage)
            .filter(LocalTrainMasterImage.image_id == image_id)
            .first()
        )

    def sync_with_harbor(self, db: Session):
        harbor_client = HarborClient(
            api_url=settings.config.registry.address,
            username=settings.config.registry.user,
            password=settings.config.registry.password,
        )
        images = harbor_client.get_master_images()
        for image in images:
            if self.get_by_image_id(db, image.image_id):
                continue
            master_image = LocalTrainMasterImage(
                registry=image.registry,
                image_id=image.image_id,
                group=image.group,
                artifact=image.artifact,
                tag=image.tag,
            )
            db.add(master_image)
            db.commit()


local_train_master_image = CRUDLocalTrainMasterImage(LocalTrainMasterImage)



from io import BytesIO
from typing import List, Union

from loguru import logger
from sqlalchemy.orm import Session
from train_lib.docker_util.docker_ops import add_archive

import docker
from docker.models.images import Image
from station.app.crud.crud_local_train import CRUDLocalTrain
from station.app.models.local_trains import LocalTrain
from station.trains.local.docker import make_docker_file


def build_train(
    db: Session,
    train_id: str,
    master_image_id: str = None,
    files: BytesIO = None,
    custom_image: str = None,
) -> str:
    if master_image_id and not files:
        raise ValueError("Must specify files with master image")

    local_crud = CRUDLocalTrain(LocalTrain)
    train = local_crud.get(db, train_id)

    image = _make_train_image(
        train_id=train_id,
        master_image=master_image_id,
        entrypoint_file=train.entrypoint,
        custom_image=custom_image,
        command=train.command,
        command_args=train.command_args,
    )

    image = _add_train_files(files=files, image=image)

    return image


def _make_train_image(
    train_id: str,
    master_image: str = None,
    entrypoint_file: str = None,
    custom_image: str = None,
    command: str = None,
    command_args: Union[List[str], str] = None,
) -> Image:
    docker_client = docker.from_env()
    if not entrypoint_file and not custom_image:
        raise ValueError(
            "Must specify an entrypoint file with master image or a custom image"
        )
    if custom_image:
        image = docker_client.images.get(custom_image)
    elif master_image:
        dockerfile = make_docker_file(
            master_image, entrypoint_file, command=command, command_args=command_args
        )
        image, logs = docker_client.images.build(fileobj=dockerfile)
        logger.info(logs)
    else:
        raise ValueError(
            "Must specify an entrypoint file with master image or a custom image"
        )

    if image.tag(repository=f"pht-local/{train_id}", tag="latest"):
        logger.debug(f"Tagged image {image.id} as pht-local/{train_id}:latest")
    else:
        raise ValueError("Failed to tag image")
    return image


def _add_train_files(files: BytesIO, image: Image) -> str:
    # add archive to container
    tag = image.tags[0]
    add_archive(tag, files, path="/opt/train")

    return tag
